#!/usr/bin/env python
# -*- coding: utf-8 -*-

from wxSchedule import wxSchedule
from wxDrawer import wxBaseDrawer, wxFancyDrawer
from wxSchedulerCore import *
import calendar
import math
import sys
import wx
import wxScheduleUtils as utils

if sys.version.startswith( "2.3" ):
	from sets import Set as set


# Events 
wxEVT_COMMAND_SCHEDULE_ACTIVATED = wx.NewEventType()
EVT_SCHEDULE_ACTIVATED = wx.PyEventBinder( wxEVT_COMMAND_SCHEDULE_ACTIVATED )

wxEVT_COMMAND_SCHEDULE_RIGHT_CLICK = wx.NewEventType()
EVT_SCHEDULE_RIGHT_CLICK = wx.PyEventBinder( wxEVT_COMMAND_SCHEDULE_RIGHT_CLICK )

wxEVT_COMMAND_SCHEDULE_DCLICK = wx.NewEventType()
EVT_SCHEDULE_DCLICK = wx.PyEventBinder( wxEVT_COMMAND_SCHEDULE_DCLICK )

wxEVT_COMMAND_PERIODWIDTH_CHANGED = wx.NewEventType()
EVT_PERIODWIDTH_CHANGED = wx.PyEventBinder( wxEVT_COMMAND_PERIODWIDTH_CHANGED )

class wxSchedulerSizer(wx.PySizer):
	def __init__(self, minSizeCallback):
		super(wxSchedulerSizer, self).__init__()

		self._minSizeCallback = minSizeCallback

	def CalcMin(self):
		return self._minSizeCallback()


# Main class
class wxSchedulerPaint( object ):
	
	def __init__( self, *args, **kwds ):
		super( wxSchedulerPaint, self ).__init__( *args, **kwds )

		self._resizable		= False
		self._style = wxSCHEDULER_VERTICAL

		self._drawerClass = wxBaseDrawer
		self._headerPanel = None

		self._schedulesCoords = list()
		self._schedulesPages = dict()

		self._datetimeCoords = []

		self._bitmap = None
		self._minSize = None
		self._drawHeaders = True
		self._guardRedraw = False

		self._periodWidth = 150
		self._headerBounds = []
		self._headerCursorState = 0
		self._headerDragOrigin = None
		self._headerDragBase = None

		# State:
		# 0 None
		# 1 Clicked, waiting for release or drag
		# 2 Dragging
		# 3 hovering up/left edge
		# 4 hovering down/right edge
		# 5 dragging up/left edige
		# 6 dragging down/right edge

		self._scheduleDraggingState = 0
		self._scheduleDragged = None
		self._scheduleDraggingOrigin = None
		self._scheduleDraggingPrevious = None
		self._scheduleDraggingStick = False

		# The highlight colour is too dark
		color = wx.SystemSettings.GetColour( wx.SYS_COLOUR_HIGHLIGHT )
		self._highlightColor = wx.Colour(int((color.Red() + 255) / 2),
						 int((color.Green() + 255) / 2),
						 int((color.Blue() + 255) / 2))

		self.pageNumber = None
		self.pageCount = 1

		if isinstance(self, wx.ScrolledWindow):
			self.SetSizer(wxSchedulerSizer(self.CalcMinSize))

	def _doClickControl( self, point, shiftDown=False ):
		if self._scheduleDraggingState in [3, 4]:
			self._scheduleDraggingState += 2
			self._scheduleDraggingOrigin = self._computeCoords(point, 0, 0)
			self._scheduleDraggingStick = shiftDown
		else:
			try:
				pMin, pMax, sch = self._findSchedule( point )
			except TypeError: # returned None
				return
			if isinstance( sch, wxSchedule ):
				self._scheduleDragged = pMin, pMax, sch
				self._scheduleDraggingState = 1
				self._scheduleDraggingOrigin = self._computeCoords(point, 0, 0)
				self._scheduleDraggingStick = shiftDown
			else:
				self._processEvt( wxEVT_COMMAND_SCHEDULE_ACTIVATED, point )

	def _doEndClickControl( self, point ):
		if self._scheduleDraggingState == 1:
			self._processEvt( wxEVT_COMMAND_SCHEDULE_ACTIVATED, point )
		elif self._scheduleDraggingState == 2:
			_, dateTime = self._computeCoords( point, 0, 0 )

			sched = self._scheduleDragged[2]
			self._drawDragging( None, self._computeAllCoords )
			sched.Offset( dateTime.Subtract( self._scheduleDraggingOrigin[1] ) )
			self._scheduleDraggingState = 0

		elif self._scheduleDraggingState in [5, 6]:
			coords = {5: self._computeStartCoords, 6: self._computeEndCoords}[self._scheduleDraggingState]
			_, _, dateTime = coords( point )

			sched = self._scheduleDragged[2]
			if self._scheduleDraggingState == 5:
				sched.SetStart( utils.copyDateTime( dateTime ) )
			else:
				sched.SetEnd( utils.copyDateTime( dateTime ) )

			self._scheduleDraggingState = 0
			self._drawDragging( None, coords )
			self.SetCursor( wx.STANDARD_CURSOR )

		self._scheduleDraggingState = 0
		self._scheduleDragged = None
		self._scheduleDraggingOrigin = None
		self._scheduleDraggingPrevious = None

	def _doMove( self, point ):
		if self._scheduleDraggingState in [0, 3, 4]:
			for sched, pointMin, pointMax in self._schedulesCoords:
				if self._style == wxSCHEDULER_VERTICAL:
					if point.x > pointMin.x and point.x < pointMax.x:
						if abs(point.y - pointMin.y) < 4:
							self._scheduleDraggingState = 3
							self.SetCursor( wx.StockCursor( wx.CURSOR_SIZENS ) )
							self._scheduleDragged = pointMin, pointMax, sched.GetClientData()
							return
						if abs(point.y - pointMax.y) < 4:
							self._scheduleDraggingState = 4
							self.SetCursor( wx.StockCursor( wx.CURSOR_SIZENS ) )
							self._scheduleDragged = pointMin, pointMax, sched.GetClientData()
							return
				else:
					if point.y > pointMin.y and point.y < pointMax.y:
						if abs(point.x - pointMin.x) < 4:
							self._scheduleDraggingState = 3
							self.SetCursor( wx.StockCursor( wx.CURSOR_SIZEWE ) )
							self._scheduleDragged = pointMin, pointMax, sched.GetClientData()
							return
						if abs(point.x - pointMax.x) < 4:
							self._scheduleDraggingState = 4
							self.SetCursor( wx.StockCursor( wx.CURSOR_SIZEWE ) )
							self._scheduleDragged = pointMin, pointMax, sched.GetClientData()
							return

			if self._scheduleDraggingState in [3, 4]:
				self._scheduleDraggingState = 0
				self.SetCursor( wx.STANDARD_CURSOR )
				self._scheduleDragged = None
		elif self._scheduleDraggingState in [5, 6]:
			self._drawDragging( point, {5: self._computeStartCoords, 6: self._computeEndCoords}[self._scheduleDraggingState] )
		elif self._scheduleDraggingState == 1:
			dx = abs(self._scheduleDraggingOrigin[0].x - point.x)
			dy = abs(self._scheduleDraggingOrigin[0].y - point.y)
			if dx >= wx.SystemSettings.GetMetric( wx.SYS_DRAG_X ) or \
			   dy >= wx.SystemSettings.GetMetric( wx.SYS_DRAG_Y ):
				self._scheduleDraggingState = 2
				self._drawDragging( point, self._computeAllCoords )
		elif self._scheduleDraggingState == 2:
			self._drawDragging( point, self._computeAllCoords )

	def _computeCoords( self, point, dx, dy ):
		pp = wx.Point(point.x + dx, point.y + dy)
		if pp.y < 0:
			pp.y = 0
		if pp.y >= self._bitmap.GetHeight():
			pp.y = self._bitmap.GetHeight() - 1
		if pp.x < 0:
			pp.x = 0
		if pp.x >= self._bitmap.GetWidth():
			pp.x = self._bitmap.GetWidth() - 1

		for idx, (dt, pointMin, pointMax) in enumerate(self._datetimeCoords):
			if pp.y >= pointMin.y and pp.y <= pointMax.y and pp.x >= pointMin.x and pp.x <= pointMax.x:
				break
		else:
			idx = -1

		if idx >= 0:
			if self._scheduleDraggingStick:
				if self._style == wxSCHEDULER_VERTICAL:
					pp = wx.Point( pp.x, pointMin.y )
				else:
					pp = wx.Point( pointMin.x, pp.y )

			theTime = utils.copyDateTime( dt )
			if self._style == wxSCHEDULER_VERTICAL:
				theTime.AddTS( wx.TimeSpan.Minutes( int(30.0 * (pp.y - pointMin.y) / (pointMax.y - pointMin.y)) ) )
			else:
				theTime.AddTS( wx.TimeSpan.Minutes( int(30.0 * (pp.x - pointMin.x) / (pointMax.x - pointMin.x)) ) )
		else:
			raise ValueError('Not found: %d %d' % (pp.x, pp.y))

		return pp, theTime

	def _computeAllCoords( self, point ):
		"""Dragging the whole schedule"""

		pMin, pMax, sch = self._scheduleDragged

		dx = point.x - self._scheduleDraggingOrigin[0].x
		dy = point.y - self._scheduleDraggingOrigin[0].y

		rMin, theTime = self._computeCoords( pMin, dx, dy )
		rMax = wx.Point( rMin.x + pMax.x - pMin.x, rMin.y + pMax.y - pMin.y )

		return rMin, rMax, theTime

	def _computeStartCoords( self, point ):
		"""Left/up resizing"""

		pMin, pMax, sch = self._scheduleDragged

 		dx = point.x - self._scheduleDraggingOrigin[0].x
 		dy = point.y - self._scheduleDraggingOrigin[0].y

		if self._style == wxSCHEDULER_VERTICAL:
			x = self._scheduleDraggingOrigin[0].x
			y = pMin.y
		else:
			x = pMin.x
			y = self._scheduleDraggingOrigin[0].y

		rMin, theTime = self._computeCoords( wx.Point(x, y), dx, dy )
		rMax = pMax

		if self._style == wxSCHEDULER_VERTICAL:
			rMin.x = pMin.x
		else:
			rMin.y = pMin.y

		return rMin, rMax, theTime

	def _computeEndCoords( self, point ):
		"""Right/down resizing"""

		pMin, pMax, sch = self._scheduleDragged

 		dx = point.x - self._scheduleDraggingOrigin[0].x
 		dy = point.y - self._scheduleDraggingOrigin[0].y

		rMin = pMin
		rMax, theTime = self._computeCoords( pMax, dx, dy )

		firstTime = self._lstDisplayedHours[0]
		lastTime = self._lstDisplayedHours[-1]

		if theTime.GetHour() == firstTime.GetHour() and \
		       theTime.GetMinute() == firstTime.GetMinute():
			theTime.SubtractTS( wx.TimeSpan.Days( 1 ) )
			theTime.SetHour( lastTime.GetHour() + 1 )
			theTime.SetMinute( 0 )

		return rMin, rMax, theTime

	def _drawDragging( self, point, coords ):
		if self._scheduleDraggingPrevious is not None:
			x, y = self.GetViewStart()
			mx, my = self.GetScrollPixelsPerUnit()
			x *= mx
			y *= my
			rMin, rMax, _ = coords( self._scheduleDraggingPrevious )
			self.RefreshRect( wx.Rect( rMin.x - x, rMin.y - y,
						   rMax.x - rMin.x, rMax.y - rMin.y ) )
			try:
				wx.Yield()
			except:
				pass

		self._scheduleDraggingPrevious = point

		if point is not None:
			rMin, rMax, _ = coords( point )

			dc = wx.ClientDC( self )
			self.PrepareDC( dc )
			dc.SetBrush( wx.Brush( self._scheduleDragged[2].GetColor() ) )
			dc.SetPen( wx.BLACK_PEN )
			dc.DrawRoundedRectangle( rMin.x, rMin.y, rMax.x - rMin.x, rMax.y - rMin.y, 5 )

	def _doRightClickControl( self, point ):
		self._processEvt( wxEVT_COMMAND_SCHEDULE_RIGHT_CLICK, point )
		
	def _doDClickControl( self, point ):
		self._processEvt( wxEVT_COMMAND_SCHEDULE_DCLICK, point )

	def _findSchedule( self, point ):
		"""
		Check if the point is on a schedule and return the schedule and its
		coordinates
		"""
		for schedule, pointMin, pointMax in self._schedulesCoords:
			inX = ( pointMin.x <= point.x ) & ( point.x <= pointMax.x )
			inY = ( pointMin.y <= point.y ) & ( point.y <= pointMax.y )
			
			if inX & inY:
				return pointMin, pointMax, schedule.GetClientData()

		for dt, pointMin, pointMax in self._datetimeCoords:
			inX = ( pointMin.x <= point.x ) & ( point.x <= pointMax.x )
			inY = ( pointMin.y <= point.y ) & ( point.y <= pointMax.y )
			
			if inX & inY:
				return pointMin, pointMax, dt


	def _getSchedInPeriod( schedules, start, end):
		"""
		Returns a list of copied schedules that intersect with
		the  period  defined by	 'start'  and 'end'.  Schedule
		start and end are trimmed so as to lie between 'start'
		and 'end'.
		"""
		results = []

		for schedule in schedules:
			try:
				schedule.bounds = None

				if schedule.start.IsLaterThan(end) or schedule.start.IsEqualTo(end):
					continue
				if start.IsLaterThan(schedule.end):
					continue

				newSchedule = schedule.Clone()
				# This is used to find the original schedule object in _findSchedule.
				newSchedule.clientdata	= schedule

				if start.IsLaterThan(schedule.start):
					newSchedule.start = utils.copyDateTime(start)
				if schedule.end.IsLaterThan(end):
					newSchedule.end = utils.copyDateTime(end)

				results.append(newSchedule)
			except wx.PyDeadObjectError:
				pass

		return results

	_getSchedInPeriod = staticmethod(_getSchedInPeriod)

	def _splitSchedules( self, schedules ):
		"""
		Returns	 a list	 of lists  of schedules.  Schedules in
		each list are guaranteed not to collide.
		"""
		results = []
		current = []

		schedules = schedules[:] # Don't alter original list

		def findNext(schedule):
			# Among schedules that start after this one ends, find the "nearest".
			candidateSchedule = None
			minDelta = None
			for sched in schedules:
				if sched.start.IsLaterThan(schedule.end) or sched.start.IsEqualTo(schedule.end):
					delta = sched.start.Subtract(schedule.end)
					if minDelta is None or minDelta > delta:
						minDelta = delta
						candidateSchedule = sched
			return candidateSchedule

		while schedules:
			schedule = schedules[0]
			while schedule:
				current.append(schedule)
				schedules.remove(schedule)
				schedule = findNext(schedule)
			results.append(current)
			current = []

		return results

	def _paintPeriod(self, drawer, start, daysCount, x, y, width, height, highlight=None):
		end = utils.copyDateTime(start)
		end.AddDS(wx.DateSpan(days=daysCount))

		blocks = self._splitSchedules(self._getSchedInPeriod(self._schedules, start, end))
		offsetY = 0

		if self._showOnlyWorkHour:
			workingHours = [(self._startingHour, self._startingPauseHour),
					(self._endingPauseHour, self._endingHour)]
		else:
			workingHours = [(self._startingHour, self._endingHour)]

		if not self.pageNumber:
			self.pageCount = 1
			self.pageLimits = [0]

			pageHeight = self.GetSize().GetHeight() - 20
			currentPageHeight = y

		for dayN in xrange(daysCount):
			theDay = utils.copyDateTime(start)
			theDay.AddDS(wx.DateSpan(days=dayN))
			theDay.SetSecond(0)
			color = highlight
			if theDay.IsSameDate( wx.DateTime.Now() ):
				if self._viewType != wxSCHEDULER_DAILY or daysCount >= 2:
					color = self._highlightColor
			drawer.DrawDayBackground( x + 1.0 * width / daysCount * dayN, y, 1.0 * width / daysCount, height,
						  highlight=color )

		if blocks:
			dayWidth = width / len(blocks)

			for idx, block in enumerate(blocks):
				maxDY = 0

				for schedule in block:
					show = True
					if self.pageNumber is not None:
						if self._schedulesPages.get(schedule.GetId(), None) != self.pageNumber:
							show = False

					if show:
						if self._style == wxSCHEDULER_VERTICAL:
							xx, yy, w, h = drawer.DrawScheduleVertical(schedule, start, workingHours,
												   x + dayWidth * idx, y,
												   dayWidth, height)
						elif self._style == wxSCHEDULER_HORIZONTAL:
							xx, yy, w, h = drawer.DrawScheduleHorizontal(schedule, start, daysCount, workingHours,
												     x, y + offsetY, width, height)
							maxDY = max(maxDY, h)

						if self.pageNumber is None:
							if currentPageHeight + h >= pageHeight:
								pageNo = self.pageCount + 1
							else:
								pageNo = self.pageCount

							self._schedulesPages[schedule.GetId()] = pageNo

						self._schedulesCoords.append((schedule, wx.Point(xx, yy), wx.Point(xx + w, yy + h)))
					else:
						schedule.Destroy()

				offsetY += maxDY

				if not self.pageNumber:
					currentPageHeight += maxDY
					if currentPageHeight >= pageHeight:
						self.pageLimits.append(currentPageHeight - maxDY)
						currentPageHeight = maxDY
						self.pageCount += 1

		now = wx.DateTime.Now()

		for dayN in xrange(daysCount):
			theDay = utils.copyDateTime(start)
			theDay.AddDS(wx.DateSpan(days=dayN))
			theDay.SetSecond(0)

			nbHours = len(self._lstDisplayedHours)

			for idx, hour in enumerate(self._lstDisplayedHours):
				theDay.SetHour(hour.GetHour())
				theDay.SetMinute(hour.GetMinute())

				if self._minSize is None or not self._resizable:
					if self._style == wxSCHEDULER_VERTICAL:
						self._datetimeCoords.append((utils.copyDateTime(theDay),
									     wx.Point(x + 1.0 * width * dayN / daysCount,
										      y + 1.0 * height * idx / nbHours),
									     wx.Point(x + 1.0 * width * (dayN + 1) / daysCount,
										      y + 1.0 * height * (idx + 1) / nbHours)))
					else:
						self._datetimeCoords.append((utils.copyDateTime(theDay),
									     wx.Point(x + 1.0 * width * (nbHours * dayN + idx) / (nbHours * daysCount),
										      y),
									     wx.Point(x + 1.0 * width * (nbHours * dayN + idx + 1) / (nbHours * daysCount),
										      y + height)))

		if isinstance(self, wx.ScrolledWindow) and self._showNow:
			now = wx.DateTime.Now()
			# This assumes self._lstDisplayedHours is sorted of course
			for dayN in xrange(daysCount):
				theDay = utils.copyDateTime(start)
				theDay.AddDS(wx.DateSpan(days=dayN))
				if theDay.IsSameDate(now):
					theDay.SetSecond(0)
					previous = None
					for idx, hour in enumerate(self._lstDisplayedHours):
						theDay.SetHour(hour.GetHour())
						theDay.SetMinute(hour.GetMinute())
						if theDay.IsLaterThan(now):
							if idx != 0:
								if self._style == wxSCHEDULER_VERTICAL:
									yPrev = y + 1.0 * height * (idx - 1) / nbHours
									delta = 1.0 * height / nbHours * now.Subtract(previous).GetSeconds() \
										/ theDay.Subtract(previous).GetSeconds()
									drawer.DrawNowHorizontal(x, yPrev + delta, width)
								else:
									xPrev = x + 1.0 * width * (nbHours * dayN + idx - 1) / (nbHours * daysCount)
									delta = 1.0 * width / (nbHours * daysCount) * now.Subtract(previous).GetSeconds() \
										/ theDay.Subtract(previous).GetSeconds()
									drawer.DrawNowVertical(xPrev + delta, y, height)
							break
						previous = utils.copyDateTime( theDay )
					break

		if self._style == wxSCHEDULER_VERTICAL:
			return max(width, DAY_SIZE_MIN.width), max(height, DAY_SIZE_MIN.height)
		else:
			return max(width, self._periodWidth), offsetY

	def _paintDay( self, drawer, day, x, y, width, height ):
		"""
		Draw column schedules
		"""

		start = utils.copyDateTime(day)
		start.SetHour(0)
		start.SetMinute(0)
		start.SetSecond(0)

		color = None
		if day.IsSameDate(wx.DateTime.Now()) and self._periodCount >= 2:
			color = self._highlightColor

		return self._paintPeriod(drawer, start, 1, x, y, width, height, highlight=color)

	def _paintDailyHeaders( self, drawer, day, x, y, width, height, includeText=True ):
		if self._style == wxSCHEDULER_HORIZONTAL:
			self._headerBounds.append((x, y, height))

		if includeText:
			color = None
			if day.IsSameDate( wx.DateTime.Now() ):
				if self._viewType != wxSCHEDULER_DAILY or self._periodCount >= 2:
					color = self._highlightColor
			w, h = drawer.DrawDayHeader(day, x, y, width, height, highlight=color)
		else:
			w, h = width, 0

		if not (self._style == wxSCHEDULER_VERTICAL or self._drawHeaders):
			hw, hh = drawer.DrawHours(x, y + h, width, height - h, self._style, includeText=includeText)
			h += hh

		return w, h

	def _paintDaily( self, drawer, day, x, y, width, height ):
		"""
		Display day schedules
		"""

		minWidth = minHeight = 0

		if self._style == wxSCHEDULER_VERTICAL:
			x += LEFT_COLUMN_SIZE
			width -= LEFT_COLUMN_SIZE

		theDay = utils.copyDateTime(day)

		if self._drawHeaders:
			maxDY = 0
			for idx in xrange(self._periodCount):
				w, h = self._paintDailyHeaders( drawer, theDay, x + 1.0 * width / self._periodCount * idx, y, 1.0 * width / self._periodCount, height)
				maxDY = max(maxDY, h)
				theDay.AddDS(wx.DateSpan(days=1))
			minHeight += maxDY
			y += maxDY
			height -= maxDY
		else:
			for idx in xrange(self._periodCount):
				self._paintDailyHeaders( drawer, theDay, x + 1.0 * width / self._periodCount * idx, y, 1.0 * width / self._periodCount, height, includeText=False )
				theDay.AddDS(wx.DateSpan(days=1))

		if self._style == wxSCHEDULER_VERTICAL:
			x -= LEFT_COLUMN_SIZE
			width += LEFT_COLUMN_SIZE

		if self._style == wxSCHEDULER_VERTICAL or self._drawHeaders:
			if self._style == wxSCHEDULER_VERTICAL:
				w, h = drawer.DrawHours(x, y, width, height, self._style)
			else:
				maxDY = 0
				for idx in xrange(self._periodCount):
					_, h = drawer.DrawHours(x + 1.0 * width / self._periodCount * idx, y, 1.0 * width / self._periodCount, height, self._style)
					maxDY = max(maxDY, h)
				w = 0
				h = maxDY
		else:
			w, h = 0, 0

		if self._style == wxSCHEDULER_VERTICAL:
			minWidth += w
			x += w
			width -= w
		else:
			minHeight += h
			y += h
			height -= h

		if self._style == wxSCHEDULER_HORIZONTAL:
			# Use directly paintPeriod or pagination fails
			w, h = self._paintPeriod(drawer, day, self._periodCount, x, y, width, height)
		else:
			w = 0
			maxDY = 0
			theDay = utils.copyDateTime(day)
			for idx in xrange(self._periodCount):
				dw, dh = self._paintDay( drawer, theDay, x + 1.0 * width / self._periodCount * idx, y, 1.0 * width / self._periodCount, height )
				w += dw
				maxDY = max(maxDY, dh)
				theDay.AddDS(wx.DateSpan(days=1))

		minWidth += w
		minHeight += h

		return minWidth, minHeight

	def _paintWeeklyHeaders( self, drawer, day, x, y, width, height ):
		firstDay = utils.setToWeekDayInSameWeek( day, 0, self._weekstart )
		firstDay.SetHour(0)
		firstDay.SetMinute(0)
		firstDay.SetSecond(0)

		maxDY = 0

		for weekday in xrange(7):
			theDay = utils.setToWeekDayInSameWeek(utils.copyDateTime(firstDay), weekday, self._weekstart)
			if theDay.IsSameDate( wx.DateTime.Now() ):
				color = self._highlightColor
			else:
				color = None
			w, h = drawer.DrawDayHeader(theDay, x + weekday * 1.0 * width / 7, y, 1.0 * width / 7, height,
						    highlight=color)
			self._headerBounds.append((int(x + (weekday + 1) * 1.0 * width / 7), y, height))
			maxDY = max(maxDY, h)

		return maxDY

	def _paintWeekly( self, drawer, day, x, y, width, height ):
		"""
		Display weekly schedule
		"""

		firstDay = utils.setToWeekDayInSameWeek( day, 0, self._weekstart )
		firstDay.SetHour(0)
		firstDay.SetMinute(0)
		firstDay.SetSecond(0)

		minWidth = minHeight = 0

		if self._style == wxSCHEDULER_VERTICAL:
			x += LEFT_COLUMN_SIZE
			width -= LEFT_COLUMN_SIZE

		maxDY = 0

		if self._drawHeaders:
			theDay = utils.copyDateTime(day)
			for idx in xrange(self._periodCount):
				maxDY = max(maxDY, self._paintWeeklyHeaders( drawer, theDay, x + 1.0 * width / self._periodCount * idx, y, 1.0 * width / self._periodCount, height ))
				theDay.AddDS(wx.DateSpan(weeks=1))

		if self._style == wxSCHEDULER_VERTICAL:
			x -= LEFT_COLUMN_SIZE
			width += LEFT_COLUMN_SIZE

		minHeight += maxDY
		y += maxDY
		height -= maxDY

		if self._style == wxSCHEDULER_VERTICAL:
			w, h = drawer.DrawHours(x, y, width, height, self._style)

			minWidth += w
			x += w
			width -= w

			day = utils.copyDateTime(firstDay)

			for idx in xrange(self._periodCount):
				for weekday in xrange(7):
					theDay = utils.setToWeekDayInSameWeek(utils.copyDateTime(day), weekday, self._weekstart)
					self._paintDay(drawer, theDay, x + (weekday + 7 * idx) * 1.0 * width / 7 / self._periodCount, y, 1.0 * width / 7 / self._periodCount, height)
				day.AddDS(wx.DateSpan(weeks=1))

			return max(WEEK_SIZE_MIN.width * self._periodCount + LEFT_COLUMN_SIZE, width), max(WEEK_SIZE_MIN.height, height)
		else:
			w, h = self._paintPeriod(drawer, firstDay, 7 * self._periodCount, x, y, width, height)

			minWidth += w
			minHeight += h

			return max(self._periodWidth * self._periodCount + LEFT_COLUMN_SIZE, minWidth), minHeight

	def _paintMonthlyHeaders( self, drawer, day, x, y, width, height ):
		if isinstance(self, wx.ScrolledWindow):
			_, h = drawer.DrawMonthHeader(day, 0, 0, self.GetSizeTuple()[0], height)
			w = width
		else:
			w, h = drawer.DrawMonthHeader(day, x, y, width, height)

		if self._style == wxSCHEDULER_HORIZONTAL:
			day.SetDay(1)
			day.SetHour(0)
			day.SetMinute(0)
			day.SetSecond(0)

			daysCount = wx.DateTime.GetNumberOfDaysInMonth(day.GetMonth())

			maxDY = 0
			for idx in xrange(daysCount):
				theDay = utils.copyDateTime(day)
				theDay.AddDS(wx.DateSpan(days=idx))
				if theDay.IsSameDate( wx.DateTime.Now() ):
					color = self._highlightColor
				else:
					color = None
				w, h = drawer.DrawSimpleDayHeader(theDay, x + 1.0 * idx * width / daysCount,
								  y + h, 1.0 * width / daysCount, height,
								  highlight=color)
				self._headerBounds.append((x + 1.0 * (idx + 1) * width / daysCount, y + h, height))
				maxDY = max(maxDY, h)

			h += maxDY

		return w, h

	def _paintMonthly( self, drawer, day, x, y, width, height):
		"""
		Draw month's calendar using calendar module functions
		"""

		if self._drawHeaders:
			w, h = self._paintMonthlyHeaders( drawer, day, x, y, width, height)
		else:
			w, h = width, 0

		y += h
		height -= h

		if self._style == wxSCHEDULER_VERTICAL:
			month = calendar.monthcalendar( day.GetYear(), day.GetMonth() + 1 )

			for w, monthWeek in enumerate( month ):
				for d, monthDay in enumerate( monthWeek ):
					cellW, cellH = 1.0 * width / 7, 1.0 * height / len(month)

					if monthDay == 0:
						theDay = None
						schedules = []
					else:
						theDay = day
						theDay.SetDay(monthDay)
						theDay.SetHour(0)
						theDay.SetMinute(0)
						theDay.SetSecond(0)

						end = utils.copyDateTime(theDay)
						end.AddDS(wx.DateSpan(days=1))

						schedules = self._getSchedInPeriod(self._schedules, theDay, end)

						if self._minSize is None or not self._resizable:
							self._datetimeCoords.append((utils.copyDateTime(theDay),
										     wx.Point(d * cellW, w * cellH),
										     wx.Point(d * cellW + cellW, w * cellH + cellH)))

					displayed = drawer.DrawSchedulesCompact(theDay, schedules, d * cellW,
										w * cellH + y, cellW, cellH,
										self._highlightColor)
					self._schedulesCoords.extend( displayed )

					for schedule in set(schedules) - set([sched for sched, _, _ in displayed]):
						schedule.Destroy()

			return (max(MONTH_CELL_SIZE_MIN.width * 7, width),
				max(MONTH_CELL_SIZE_MIN.height * (w + 1), height))
		else:
			day.SetDay(1)
			day.SetHour(0)
			day.SetMinute(0)
			day.SetSecond(0)

			daysCount = wx.DateTime.GetNumberOfDaysInMonth(day.GetMonth())

			minHeight = h

			w, h = self._paintPeriod(drawer, day, daysCount, x, y, width, height)
			minHeight += h

			return w, minHeight

	def _processEvt( self, commandEvent, point ):
		""" 
		Process the command event passed at the given point
		"""
		evt = wx.PyCommandEvent( commandEvent )
		_, _, sch = self._findSchedule( point )
		if isinstance( sch, wxSchedule ):
			mySch = sch
			myDate = None
		else:
			mySch = None
			myDate = sch
		
		evt.schedule = mySch
		evt.date = myDate
		evt.SetEventObject( self )
		self.ProcessEvent( evt ) 

	def DoPaint(self, drawer, x, y, width, height):
		for schedule, _, _ in self._schedulesCoords:
			schedule.Destroy()

		self._schedulesCoords = list()

		day = utils.copyDate(self.GetDate())

		if self._viewType == wxSCHEDULER_DAILY:
			return self._paintDaily(drawer, day, x, y, width, height)
		elif self._viewType == wxSCHEDULER_WEEKLY:
			return self._paintWeekly(drawer, day, x, y, width, height)
		elif self._viewType == wxSCHEDULER_MONTHLY:
			return self._paintMonthly(drawer, day, x, y, width, height)

	def GetViewSize(self):
		# Used by wxSchedulerReport

		size = self.GetSize()
		minSize = self.CalcMinSize()

		return wx.Size(max(size.width, minSize.width), max(size.height, minSize.height))

	def _CalcMinSize(self):
		if self._viewType == wxSCHEDULER_DAILY:
			minW, minH = DAY_SIZE_MIN.width, DAY_SIZE_MIN.height
		elif self._viewType == wxSCHEDULER_WEEKLY:
			minW, minH = WEEK_SIZE_MIN.width, WEEK_SIZE_MIN.height
		elif self._viewType == wxSCHEDULER_MONTHLY:
			minW, minH = MONTH_CELL_SIZE_MIN.width * 7, 0 # will be computed

		if self._viewType == wxSCHEDULER_MONTHLY or self._style == wxSCHEDULER_HORIZONTAL:
			memDC = wx.MemoryDC()
			bmp = wx.EmptyBitmap(1, 1)
			memDC.SelectObject(bmp)
			try:
				if self._drawerClass.use_gc:
					context = wx.GraphicsContext.Create(memDC)
					context.SetFont(wx.NORMAL_FONT, wx.BLACK)
				else:
					context = memDC
					context.SetFont(wx.NORMAL_FONT)

				if isinstance(self, wx.ScrolledWindow):
					size = self.GetVirtualSize()
				else:
					size = self.GetSize()

				# Actually, only the min height may vary...
				_, minH = self.DoPaint(self._drawerClass(context, self._lstDisplayedHours),
						       0, 0, size.GetWidth(), 0)

				if self._style == wxSCHEDULER_HORIZONTAL:
					if self._viewType == wxSCHEDULER_DAILY:
						minW = self._periodWidth * 4
					elif self._viewType == wxSCHEDULER_WEEKLY:
						minW = self._periodWidth * 7
					elif self._viewType == wxSCHEDULER_MONTHLY:
						return wx.Size(self._periodWidth * wx.DateTime.GetNumberOfDaysInMonth(self.GetDate().GetMonth()), minH)
				elif self._viewType == wxSCHEDULER_MONTHLY:
					return wx.Size(minW, minH)
			finally:
				memDC.SelectObject(wx.NullBitmap)
		elif self._style == wxSCHEDULER_VERTICAL:
			return wx.Size(minW * self._periodCount + LEFT_COLUMN_SIZE, minH)

		return wx.Size(minW * self._periodCount, minH)

	def CalcMinSize(self):
		if self._minSize is None:
			self._minSize = self._CalcMinSize()
		return self._minSize

	def InvalidateMinSize(self):
		self._minSize = None
		self._datetimeCoords = list()

	def DrawBuffer( self ):
		if isinstance(self, wx.ScrolledWindow):
			if self._resizable:
				size = self.GetVirtualSize()
			else:
				size = self.CalcMinSize()
		else:
			size = self.GetSize()

		self._bitmap = wx.EmptyBitmap(size.GetWidth(), size.GetHeight())
		memDC = wx.MemoryDC()
		memDC.SelectObject(self._bitmap)
		try:
			memDC.BeginDrawing()
			try:
				memDC.SetBackground( wx.Brush( SCHEDULER_BACKGROUND_BRUSH() ) )
				memDC.SetPen( FOREGROUND_PEN )
				memDC.Clear()
				memDC.SetFont(wx.NORMAL_FONT)

				if self._drawerClass.use_gc:
					context = wx.GraphicsContext.Create(memDC)
					context.SetFont(wx.NORMAL_FONT, wx.BLACK)
				else:
					context = memDC
					context.SetFont(wx.NORMAL_FONT)

				width, height = self.DoPaint(self._drawerClass(context, self._lstDisplayedHours),
							     0, 0, size.GetWidth(), size.GetHeight())
			finally:
				memDC.EndDrawing()
		finally:
			memDC.SelectObject(wx.NullBitmap)

		# Bad things may happen here from time to time.
		if isinstance(self, wx.ScrolledWindow):
			if self._resizable and not self._guardRedraw:
				self._guardRedraw = True
				try:
					if int(width) > size.GetWidth() or int(height) > size.GetHeight():
						self.SetVirtualSize(wx.Size(int(width), int(height)))
						self.DrawBuffer()
				finally:
					self._guardRedraw = False

	def RefreshSchedule( self, schedule ):
		if schedule.bounds is not None:
			memDC = wx.MemoryDC()
			memDC.SelectObject(self._bitmap)
			try:
				memDC.BeginDrawing()
				memDC.SetBackground( wx.Brush( SCHEDULER_BACKGROUND_BRUSH() ) )
				memDC.SetPen( FOREGROUND_PEN )
				memDC.SetFont(wx.NORMAL_FONT)

				if self._drawerClass.use_gc:
					context = wx.GraphicsContext.Create(memDC)
					context.SetFont(wx.NORMAL_FONT, wx.BLACK)
				else:
					context = memDC
					context.SetFont(wx.NORMAL_FONT)

				self._drawerClass(context, self._lstDisplayedHours)._DrawSchedule(schedule, *schedule.bounds)
			finally:
				memDC.SelectObject(wx.NullBitmap)

			originX, originY = self.GetViewStart()
			unitX, unitY = self.GetScrollPixelsPerUnit()
			x, y, w, h = schedule.bounds
			self.RefreshRect(wx.Rect(math.floor(x - originX * unitX) - 1, math.floor(y - originY * unitY) - 1,
						 math.ceil(w) + 2, math.ceil(h) + 2))

	def OnPaint( self, evt = None ):
		# Do the draw

		if self._dc:
			dc = self._dc
		else:
			dc = wx.PaintDC(self)
			self.PrepareDC(dc)

		dc.BeginDrawing()
		try:
			dc.DrawBitmap(self._bitmap, 0, 0, False)
		finally:
			dc.EndDrawing()

	def SetResizable( self, value ):
		"""
		Draw proportionally of actual space but not down on minimun sizes
		The actual sze is retrieved by GetSize() method of derived object
		"""
		self._resizable = bool( value )

	def SetStyle(self, style):
		"""
		Sets  the drawing  style.  Values  for 'style'	may be
		wxSCHEDULER_VERTICAL	   (the	      default)	    or
		wxSCHEDULER_HORIZONTAL.
		"""
		self._style = style
		self.InvalidateMinSize()
		self.Refresh()

	def GetStyle( self ):
		"""
		Returns the current drawing style.
		"""
		return self._style

	def SetHighlightColor( self, color ):
		"""
		Sets the highlight color, i.e. the color used to draw
		today's background.
		"""

		self._highlightColor = color

	def GetHighlightColor( self ):
		"""
		Returns the current highlight color.
		"""

		return self._highlightColor

	def SetDrawer(self, drawerClass):
		"""
		Sets the drawer class.
		"""
		self._drawerClass = drawerClass
		self.InvalidateMinSize()
		self.Refresh()

	def GetDrawer(self):
		return self._drawerClass

	def SetPeriodWidth( self, width ):
		"""Sets the width of a day in horizontal mode."""

		self._periodWidth = width

	def GetPeriodWidth( self ):
		return self._periodWidth

	def Refresh( self ):
		self.DrawBuffer()
		super( wxSchedulerPaint, self ).Refresh()
		if self._headerPanel is not None:
			self._headerPanel.Refresh()

	def SetHeaderPanel( self, panel ):
		"""
		Call this with an instance of wx.Panel. The headers
		will then be painted on this panel, and thus will be
		unaffected by vertical scrolling. The panel will be
		resized as needed.
		"""

		self._drawHeaders = False
		self._headerPanel = panel

		panel.Bind( wx.EVT_PAINT, self._OnPaintHeaders )
		panel.Bind( wx.EVT_MOTION, self._OnMoveHeaders )
		panel.Bind( wx.EVT_LEAVE_WINDOW, self._OnLeaveHeaders )
		panel.Bind( wx.EVT_LEFT_DOWN, self._OnLeftDownHeaders )
		panel.Bind( wx.EVT_LEFT_UP, self._OnLeftUpHeaders )
		panel.SetSize(wx.Size(-1, 1))
		self.Bind(wx.EVT_SCROLLWIN, self._OnScroll)

		panel.Refresh()

	# Headers stuff

	def _OnPaintHeaders( self, evt ):
		dc = wx.PaintDC( self._headerPanel )
		dc.BeginDrawing()
		try:
			dc.SetBackground( wx.Brush( SCHEDULER_BACKGROUND_BRUSH() ) )
			dc.SetPen( FOREGROUND_PEN )
			dc.Clear()
			dc.SetFont( wx.NORMAL_FONT )

			if self._drawerClass.use_gc:
				context = wx.GraphicsContext.Create(dc)
				context.SetFont(wx.NORMAL_FONT, wx.BLACK)
			else:
				context = dc
				context.SetFont(wx.NORMAL_FONT)

			drawer = self._drawerClass(context, self._lstDisplayedHours)

			if self._resizable:
				width, _ = self.GetVirtualSize()
			else:
				width, _ = self.CalcMinSize()

			day = utils.copyDate(self.GetDate())

			x, y = 0, 0

			# Take horizontal scrolling into account
			x0, _ = self.GetViewStart()
			xu, _ = self.GetScrollPixelsPerUnit()
			x0 *= xu
			x -= x0

			self._headerBounds = []

			if self._viewType == wxSCHEDULER_DAILY:
				if self._style == wxSCHEDULER_VERTICAL:
					x += LEFT_COLUMN_SIZE
					width -= LEFT_COLUMN_SIZE
				theDay = utils.copyDateTime(day)
				maxDY = 0
				for idx in xrange(self._periodCount):
					_, h = self._paintDailyHeaders(drawer, theDay, x + 1.0 * width / self._periodCount * idx, y, 1.0 * width / self._periodCount, 36)
					maxDY = max(maxDY, h)
					theDay.AddDS(wx.DateSpan(days=1))
				h = maxDY
			elif self._viewType == wxSCHEDULER_WEEKLY:
				if self._style == wxSCHEDULER_VERTICAL:
					x += LEFT_COLUMN_SIZE
					width -= LEFT_COLUMN_SIZE
				theDay = utils.copyDateTime(day)
				maxDY = 0
				for idx in xrange(self._periodCount):
					h = self._paintWeeklyHeaders(drawer, theDay, x + 1.0 * width / self._periodCount * idx, y, 1.0 * width / self._periodCount, 36)
					maxDY = max(maxDY, h)
					theDay.AddDS(wx.DateSpan(weeks=1))
				h = maxDY
			elif self._viewType == wxSCHEDULER_MONTHLY:
				_, h = self._paintMonthlyHeaders(drawer, day, x, y, width, 36)

			minW, minH = self._headerPanel.GetMinSize()
			if minH != h:
				self._headerPanel.SetMinSize(wx.Size(-1, h))
				self._headerPanel.GetParent().Layout()

			# Mmmmh, maybe we'll support this later, but not right now
			if self._style == wxSCHEDULER_VERTICAL:
				self._headerBounds = []
		finally:
			dc.EndDrawing()

	def _OnMoveHeaders( self, evt ):
		if self._headerDragOrigin is None:
			for x, y, h in self._headerBounds:
				if abs(evt.GetX() - x) < 5 and evt.GetY() >= y and evt.GetY() < y + h:
					if self._headerCursorState == 0:
						self._headerCursorState = 1
						self._headerPanel.SetCursor( wx.StockCursor( wx.CURSOR_SIZEWE ) )
					break
			else:
				if self._headerCursorState == 1:
					self._headerPanel.SetCursor( wx.STANDARD_CURSOR )
					self._headerCursorState = 0
		else:
			deltaX = evt.GetX() - self._headerDragOrigin
			self.SetPeriodWidth( max(50, self._headerDragBase + deltaX) )

			evt = wx.PyCommandEvent( wxEVT_COMMAND_PERIODWIDTH_CHANGED )
			evt.SetEventObject( self )
			self.ProcessEvent( evt ) 

			self.InvalidateMinSize()
			self.Refresh()

	def _OnLeaveHeaders( self, evt ):
		if self._headerCursorState == 1 and self._headerDragOrigin is None:
			self._headerPanel.SetCursor( wx.STANDARD_CURSOR )
			self._headerCursorState = 0

	def _OnLeftDownHeaders( self, evt ):
		if self._headerCursorState == 1:
			self._headerDragOrigin = evt.GetX()
			self._headerDragBase = self._periodWidth
			self._headerPanel.CaptureMouse()
		else:
			evt.Skip()

	def _OnLeftUpHeaders( self, evt ):
		if self._headerCursorState == 1:
			self._headerPanel.ReleaseMouse()
			self._headerDragOrigin = None
		else:
			evt.Skip()

	def _OnScroll( self, evt ):
		self._headerPanel.Refresh()
		evt.Skip()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.i18n import _
from taskcoachlib import meta

# Help texts for menu items and toolbar buttons (so called uicommands). 
# At the moment this lists mostly the uicommands that have a keyboard shortcut 
# because we need to be able to reuse the help text in the help dialog.

addAttachment = _('Browse for files to add as attachment to the selected item(s)')
addNote = _('Add a note to the selected item(s)')
categoryNew = _('Insert a new category')
editCut = _('Cut the selected item(s) to the clipboard')
editCopy = _('Copy the selected item(s) to the clipboard')
editPaste = _('Paste item(s) from the clipboard')
editPasteAsSubitem = _('Paste item(s) from the clipboard as subitem of the selected item')
editPreferences = _('Edit preferences')
editRedo = _('Redo the last command that was undone')
editSelectAll = _('Select all items in the current view')
editUndo = _('Undo the last command')
effortNew = _('Add an effort period to the selected task(s)')
effortStart = _('Start tracking effort for the selected task(s)')
effortStopOrResume = _('Stop tracking effort or resume tracking effort')
fileClose = _('Close the current file')
fileQuit = _('Exit %s')%meta.name
fileOpen = _('Open a %s file')%meta.name
fileSave = _('Save the current file')
fileMergeDiskChanges = _('Load what has changed on disk')
fileSaveAs = _('Save the current file under a new name')
help = _('Help about the program') # pylint: disable=W0622
mailItem = _('Mail the selected item(s), using your default mailer')
noteNew = _('Insert a new note')
openAllAttachments = _('Open all attachments of the selected item(s)')
openAllNotes = _('Open all notes of the selected item(s)')
print_ = _('Print the current file')
printPageSetup = _('Setup the characteristics of the printer page')
resetFilter = _('Show all items (reset all filters)')
resetCategoryFilter = _('Show all items regardless of category')
search = _('Move keyboard focus from viewer to search control')
taskDecreasePriority = _('Decrease the priority of the selected task(s)')
taskIncreasePriority = _('Increase the priority of the selected task(s)')
taskMaxPriority = _('Make the selected task(s) the highest priority task(s)')
taskMinPriority = _('Make the selected task(s) the lowest priority task(s)')
taskNew = _('Insert a new task')
viewCollapseAll = _('Collapse all items with subitems')
viewExpandAll = _('Expand all items with subitems')
viewNextViewer = _('Activate the next open viewer')
viewPreviousViewer = _('Activate the previous open viewer')



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import warnings
import wx
import time

from wxScheduleUtils import copyDateTime

#New event
wxEVT_COMMAND_SCHEDULE_CHANGE = wx.NewEventType()
EVT_SCHEDULE_CHANGE = wx.PyEventBinder( wxEVT_COMMAND_SCHEDULE_CHANGE )


# Constants
			 
			 
class wxSchedule( wx.EvtHandler ):
	
	SCHEDULE_DEFAULT_COLOR = wx.Colour( 247, 212, 57 )  
	SCHEDULE_DEFAULT_FOREGROUND = wx.BLACK

	CATEGORIES = {
		"Work"		: wx.GREEN,
		"Holiday"	: wx.GREEN,
		"Phone"		: wx.GREEN,
		"Email"		: wx.GREEN,
		"Birthday"	: wx.GREEN,
		"Ill"		: wx.GREEN,
		"At home"	: wx.GREEN,
		"Fax"		: wx.GREEN, 
	}
	
	def __init__( self ):
		"""
		Use self.start and self.end for set the star and the end of the schedule.
		If both start and end datetime have time set to 00:00 the schedule is
		relative on entire day/days.
		"""
		super( wxSchedule, self ).__init__()
		
		self._color			= self.SCHEDULE_DEFAULT_COLOR
		self._font                      = wx.NORMAL_FONT
		self._foreground                = self.SCHEDULE_DEFAULT_FOREGROUND
		self._category		= "Work"
		self._description	= ''
		self._notes			= ''
		self._end			= wx.DateTime().Now()
		self._start			= wx.DateTime().Now()
		self._done			= False
		self._clientdata	= None
		self._icons			= []
		self._complete = None
		self._id = '%.f-%s' % (time.time(), id(self))
		
		# Need for freeze the event notification
		self._freeze = False
		self._layoutNeeded = False
	
	def __getattr__(self, name):
		if name[:3] in [ 'get', 'set' ]:
			warnings.warn( "getData() is deprecated, use GetData() instead", DeprecationWarning, stacklevel=2 )
			
			name = name[0].upper() + name[1:]
		
			return getattr(self, name)

		raise AttributeError(name)
		
	# Global methods
	def Freeze( self ):
		# Freeze the event notification
		self._freeze = True
		self._layoutNeeded = False
	
	def Thaw( self ):
		# Wake up the event
		self._freeze = False
		self._eventNotification( self._layoutNeeded )

	def GetData(self):
		"""
		Return wxSchedule data into a dict
		"""
		attributes = [ 
			"category", 
			"color",
			"font",
			"foreground",
			"description", 
			"done", 
			"end", 
			"notes",
			"start", 
			"clientdata",
			"icons",
			"complete",
			"id",
		]
		data = dict()
					 
		for attribute in attributes:
			data[ attribute ] = self.__getattribute__( attribute )
		
		return data

	def Clone(self):
		newSchedule = wxSchedule()
		for name, value in self.GetData().items():
			setattr(newSchedule, name, value)
		# start and end should be copied as well
		newSchedule._start = copyDateTime(newSchedule._start)
		newSchedule._end = copyDateTime(newSchedule._end)
		return newSchedule

	# Internal methods
	
	def _eventNotification( self, layoutNeeded=False ):
		""" If not freeze, wake up and call the event notification
		"""
		if self._freeze:
			self._layoutNeeded = self._layoutNeeded or layoutNeeded
			return

		#Create the event and propagete it
		evt = wx.PyCommandEvent( wxEVT_COMMAND_SCHEDULE_CHANGE )
		
		evt.category	= self._category
		evt.color		= self._color
		evt.font                = self._font
		evt.foreground = self._foreground
		evt.description	= self._description
		evt.done		= self._done
		evt.end			= self._end
		evt.notes		= self._notes
		evt.start		= self._start
		evt.icons		= self._icons
		evt.complete            = self._complete
		evt.schedule	= self
		evt.layoutNeeded = layoutNeeded

		evt.SetEventObject( self )
		
		self.ProcessEvent( evt )

	def __eq__( self, schedule ):
		"""
		Control if the schedule passed are equal than me
		"""
		# Is not a wxSchedule
		if not isinstance( schedule, wxSchedule ): 
			return False
		
		# Check wxSchedules attributes
		return self.GetData() == schedule.GetData()
		
	# Properties
	def SetCategory( self, category ):
		"""
		Set the color
		"""
		if category not in self.CATEGORIES.keys():
			raise ValueError, "%s is not a valid category" % category
		
		self._category = category
		self._eventNotification()
	
	def GetCategory( self ):
		""" 
		Return the current category
		"""
		return self._category
	
	def SetColor( self, color ):
		"""
		Set the color
		"""
		if not isinstance( color, wx.Colour ):
			raise ValueError, "Color can be only a wx.Colour value"

		self._color = color
		self._eventNotification()
		
	def GetColor( self ):
		""" 
		Return the color
		"""
		return self._color

	def SetFont( self, font ):
		"""
		Set the font
		"""

		if font is None:
			self._font = wx.NORMAL_FONT
		else:
			self._font = font

		self._eventNotification()

	def GetFont( self ):
		"""
		Return the font
		"""
		return self._font

	def SetForeground( self, color ):
		"""
		Sets the text color
		"""
		self._foreground = color

	def GetForeground( self ):
		"""
		Returns the text color
		"""
		return self._foreground

	def SetDescription( self, description ):
		"""
		Set the description
		"""
		if not isinstance( description, basestring ):
			raise ValueError, "Description can be only a str value"

		self._description = description
		self._eventNotification( True )
		
	def GetDescription( self ):
		"""
		Return the description
		"""
		return self._description

	def SetDone( self, done ):
		""" 
		Are this schedule complete?
		""" 
		if not isinstance( done, bool ):
			raise ValueError, "Done can be only a bool value"
		
		self._done = done
		self._eventNotification()
		
	def GetDone( self ):
		"""
		Return the done value
		"""
		return self._done
	
	def SetEnd( self, dtEnd ):
		"""
		Set the end
		"""
		if not isinstance( dtEnd, wx.DateTime ):
			raise ValueError, "dateTime can be only a wx.DateTime value"

		self._end = dtEnd
		self._eventNotification( True )
	
	def GetEnd( self ):
		""" 
		Return the end
		"""
		return self._end
				
	def SetNotes( self, notes ):
		""" 
		Set the notes
		"""
		if not isinstance( notes, basestring ):
			raise ValueError, "notes can be only a str value"
	   
		self._notes = notes
		self._eventNotification()

	def GetNotes( self ):
		""" 
		Return the notes
		"""
		return self._notes

	def SetStart( self, dtStart ):
		""" Set the start
		"""
		if not isinstance( dtStart, wx.DateTime ):
			raise ValueError, "dateTime can be only a wx.DateTime value"
		
		self._start = dtStart
		self._eventNotification( True )
	
	def GetStart( self ):
		""" 
		Return the start
		"""
		return self._start

	def Offset( self, ts ):
		"""
		Offsets the schedule by the specified time span.
		"""
		self._start.AddTS( ts )
		self._end.AddTS( ts )
		self._eventNotification( True )

	def GetIcons(self):
		return self._icons
	
	def SetIcons(self, icons):
		layoutNeeded = (bool(icons) and not bool(self._icons)) or \
			       (bool(self._icons) and not bool(icons))
		self._icons = icons
		
		self._eventNotification( layoutNeeded )

	def GetComplete(self):
		return self._complete

	def SetComplete(self, complete):
		layoutNeeded = (self._complete is None and complete is not None) or \
			       (self._complete is not None and complete is None)
		self._complete = complete
		self._eventNotification( layoutNeeded )

	def SetClientData( self, clientdata ):
		self._clientdata = clientdata
	
	def GetClientData( self ):
		return self._clientdata

	def SetId( self, id_ ):
		self._id = id_

	def GetId( self ):
		return self._id

	category = property( GetCategory, SetCategory )
	color = property( GetColor, SetColor )
	font = property( GetFont, SetFont )
	foreground = property( GetForeground, SetForeground )
	description = property( GetDescription, SetDescription )
	done = property( GetDone, SetDone )
	start = property( GetStart, SetStart )
	end = property( GetEnd, SetEnd )
	notes = property( GetNotes, SetNotes )
	clientdata = property( GetClientData, SetClientData )
	icons = property( GetIcons, SetIcons )
	complete = property( GetComplete, SetComplete )
	id = property( GetId, SetId )



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import re, shutil, os
from taskcoachlib.domain import date


class TodoTxtWriter(object):
    VERSION = 1

    def __init__(self, fd, filename):
        self.__fd = fd
        self.__filename = filename
        self.__maxDateTime = date.DateTime()
        
    def write(self, viewer, settings, selectionOnly, **kwargs):
        tasks = viewer.visibleItems()
        if selectionOnly:
            tasks = [task for task in tasks if viewer.isselected(task)]
        return self.writeTasks(tasks)
    
    def writeTasks(self, tasks):
        count = 0
        for task in tasks:
            count += 1
            self.__fd.write(self.priority(task.priority()) + \
                            self.completionDate(task.completionDateTime()) + \
                            self.startDate(task.plannedStartDateTime()) + \
                            task.subject(recursive=True) + \
                            self.contextsAndProjects(task) + \
                            self.dueDate(task.dueDateTime()) + \
                            self.id(task.id()) + '\n')
        metaName = self.__filename + '-meta'
        if os.path.exists(metaName):
            os.remove(metaName)
        if os.path.exists(self.__filename): # Unit tests
            self.__fd.close()
            with file(metaName, 'wb') as dst:
                dst.write('VERSION: %d\n' % self.VERSION)
                with file(self.__filename, 'rb') as src:
                    shutil.copyfileobj(src, dst)
        return count

    @staticmethod
    def priority(priorityNumber):
        return '(%s) '%chr(ord('A') + priorityNumber - 1) if 1 <= priorityNumber <= 26 else ''

    @classmethod
    def startDate(cls, plannedStartDateTime):
        return '%s '%cls.dateTime(plannedStartDateTime) if cls.isActualDateTime(plannedStartDateTime) else ''
    
    @classmethod
    def dueDate(cls, dueDateTime):
        return ' due:%s'%cls.dateTime(dueDateTime) if cls.isActualDateTime(dueDateTime) else ''

    @classmethod
    def id(cls, id_):
        return ' tcid:%s' % id_

    @classmethod
    def completionDate(cls, completionDateTime):
        return 'X ' + '%s '%cls.dateTime(completionDateTime) if cls.isActualDateTime(completionDateTime) else ''
        
    @staticmethod
    def dateTime(dateTime):
        ''' Todo.txt doesn't support time, just dates, so ignore the time part. '''
        return dateTime.date().strftime('%Y-%m-%d')

    @staticmethod
    def isActualDateTime(dateTime, maxDateTime=date.DateTime()):
        return dateTime != maxDateTime

    @classmethod
    def contextsAndProjects(cls, task):
        subjects = []
        for category in task.categories():
            subject = category.subject(recursive=True).strip()
            if subject and subject[0] in ('@', '+'):
                subject = re.sub(r' -> ', '->', subject)
                subject = re.sub(r'\s+', '_', subject)
                subjects.append(subject)
        return ' ' + ' '.join(sorted(subjects)) if subjects else ''



# -*- coding: utf-8 -*-

'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import render
from taskcoachlib.i18n import _
from taskcoachlib.domain import date


class RenderDateTime(test.TestCase):
    def assertRenderedDateTime(self, expectedDateTime, *dateTimeArgs):
        renderedDateTime = render.dateTime(date.DateTime(*dateTimeArgs))
        if expectedDateTime:
            renderedParts = renderedDateTime.split(' ', 1)
            if len(renderedParts) > 1:
                renderedDate, renderedTime = renderedParts
                expectedDate, expectedTime = expectedDateTime.split(' ', 1)
                self.assertEqual(expectedTime, renderedTime)
            else:
                expectedDate, renderedDate = expectedDateTime, renderedDateTime
            self.assertEqual(expectedDate, renderedDate)
        else:
            self.assertEqual(expectedDateTime, renderedDateTime)
            
    @staticmethod
    def expectedDateTime(*dateTimeArgs):
        return render.dateTimeFunc(date.DateTime(*dateTimeArgs))

    @staticmethod
    def expectedDate(*dateTimeArgs):
        return render.dateFunc(date.DateTime(*dateTimeArgs))
        
    def testSomeRandomDateTime(self):
        expectedDateTime = self.expectedDateTime(2010, 4, 5, 12, 54)
        self.assertRenderedDateTime(expectedDateTime, 2010, 4, 5, 12, 54, 42)
        
    def testInfiniteDateTime(self):
        self.assertRenderedDateTime('')
        
    def testStartOfDay(self):
        expectedDateTime = self.expectedDate(2010, 4, 5)
        self.assertRenderedDateTime(expectedDateTime, 2010, 4, 5)

    def testEndOfDay(self):
        expectedDateTime = self.expectedDate(2010, 4, 5)
        self.assertRenderedDateTime(expectedDateTime, 2010, 4, 5, 23, 59, 59)

    def testEndOfDayWithoutSeconds(self):
        expectedDateTime = self.expectedDate(2010, 4, 5)
        self.assertRenderedDateTime(expectedDateTime, 2010, 4, 5, 23, 59)

    def testAlmostStartOfDay(self):
        expectedDateTime = self.expectedDateTime(2010, 4, 5, 0, 1)
        self.assertRenderedDateTime(expectedDateTime, 2010, 4, 5, 0, 1, 0)

    def testAlmostEndOfDay(self):
        expectedDateTime = self.expectedDateTime(2010, 4, 5, 23, 58)
        self.assertRenderedDateTime(expectedDateTime, 2010, 4, 5, 23, 58, 59)

    def testElevenOClock(self):
        expectedDateTime = self.expectedDateTime(2010, 4, 5, 23, 0)
        self.assertRenderedDateTime(expectedDateTime, 2010, 4, 5, 23, 0, 0)
        
    def testDateBefore1900(self):
        # Don't check for '1801' since the year may be formatted on only 2
        # digits.
        result = render.dateTime(date.DateTime(1801, 4, 5, 23, 0, 0))
        self.failUnless('01' in result, result)
                         
                         
class RenderDate(test.TestCase):
    def testRenderDateWithDateTime(self):
        self.assertEqual(render.date(date.DateTime(2000, 1, 1)), 
                         render.date(date.DateTime(2000, 1, 1, 10, 11, 12)))
        

class RenderTimeLeftTest(test.TestCase):
    def testNoTimeLeftWhenActive(self):
        timeLeft = date.TimeDelta()
        self.assertEqual('0:00', render.timeLeft(timeLeft, False))

    def testNoTimeLeftWhenCompleted(self):
        self.assertEqual('', render.timeLeft(date.TimeDelta(), True))

    def testNoTimeLeftWhenNoDueDate(self):
        self.assertEqual('', render.timeLeft(date.TimeDelta.max, False))

    def testInfiniteTimeLeftWhenCompleted(self):
        self.assertEqual('', render.timeLeft(date.TimeDelta.max, True))

    def testOneDayLeftWhenActive(self):
        timeLeft = date.TimeDelta(days=1)
        self.assertEqual('1 day, 0:00', render.timeLeft(timeLeft, False))

    def testOneDayLeftWhenCompleted(self):
        timeLeft = date.TimeDelta(days=1)
        self.assertEqual('', render.timeLeft(timeLeft, True))

    def testTwoDaysLeftWhenActive(self):
        timeLeft = date.TimeDelta(days=2)
        self.assertEqual('2 days, 0:00', render.timeLeft(timeLeft, False))

    def testTwoDaysLeftWhenCompleted(self):
        timeLeft = date.TimeDelta(days=2)
        self.assertEqual('', render.timeLeft(timeLeft, True))

    def testOneDayLateWhenActive(self):
        timeLeft = date.TimeDelta(days=-1)
        self.assertEqual('-1 day, 0:00', render.timeLeft(timeLeft, False))

    def testOneDayLateWhenCompleted(self):
        timeLeft = date.TimeDelta(days=-1)
        self.assertEqual('', render.timeLeft(timeLeft, True))

    def testOneHourLateWhenActive(self):
        timeLeft = -date.ONE_HOUR
        self.assertEqual('-1:00', render.timeLeft(timeLeft, False))

    def testOneDayHourWhenCompleted(self):
        timeLeft = -date.ONE_HOUR
        self.assertEqual('', render.timeLeft(timeLeft, True))


class RenderTimeSpentTest(test.TestCase):
    def testZeroTime(self):
        self.assertEqual('', render.timeSpent(date.TimeDelta()))
        
    def testOneSecond(self):
        self.assertEqual('0:00:01', render.timeSpent(date.ONE_SECOND))
            
    def testTenHours(self):
        self.assertEqual('10:00:00', 
            render.timeSpent(date.TimeDelta(hours=10)))
            
    def testNegativeHours(self):
        self.assertEqual('-1:00:00', 
                         render.timeSpent(date.TimeDelta(hours=-1)))
        
    def testNegativeSeconds(self):
        self.assertEqual('-0:00:01', 
                         render.timeSpent(date.TimeDelta(seconds=-1)))

    def testDecimalNegative(self):
        self.assertEqual('-1.25', render.timeSpent(date.TimeDelta(hours=-1, minutes=-15), decimal=True))

    def testDecimalNul(self):
        self.assertEqual('', render.timeSpent(date.TimeDelta(hours=0), decimal=True))

    def testDecimal(self):
        self.assertEqual('0.50', render.timeSpent(date.TimeDelta(minutes=30), decimal=True))


class RenderWeekNumberTest(test.TestCase):
    def testWeek1(self):
        self.assertEqual('2005-1', 
                         render.weekNumber(date.DateTime(2005, 1, 3)))
        
    def testWeek53(self):
        self.assertEqual('2004-53', 
                         render.weekNumber(date.DateTime(2004, 12, 31)))
        
        
class RenderRecurrenceTest(test.TestCase):
    def testNoRecurrence(self):
        self.assertEqual('', render.recurrence(date.Recurrence()))
        
    def testDailyRecurrence(self):
        self.assertEqual(_('Daily'), 
                         render.recurrence(date.Recurrence('daily')))
        
    def testWeeklyRecurrence(self):
        self.assertEqual(_('Weekly'), 
                         render.recurrence(date.Recurrence('weekly')))
        
    def testMonthlyRecurrence(self):
        self.assertEqual(_('Monthly'), 
                         render.recurrence(date.Recurrence('monthly')))

    def testYearlyRecurrence(self):
        self.assertEqual(_('Yearly'), 
                         render.recurrence(date.Recurrence('yearly')))

    def testEveryOtherDay(self):
        self.assertEqual(_('Every other day'), 
                         render.recurrence(date.Recurrence('daily', amount=2)))
        
    def testEveryOtherWeek(self):
        self.assertEqual(_('Every other week'), 
                         render.recurrence(date.Recurrence('weekly', amount=2)))
        
    def testEveryOtherMonth(self):
        self.assertEqual(_('Every other month'), 
                         render.recurrence(date.Recurrence('monthly', amount=2)))
        
    def testEveryOtherYear(self):
        self.assertEqual(_('Every other year'), 
                         render.recurrence(date.Recurrence('yearly', amount=2)))
        
    def testThreeDaily(self):
        self.assertEqual('Every 3 days', 
                         render.recurrence(date.Recurrence('daily', amount=3))) 
        
    def testThreeWeekly(self):
        self.assertEqual('Every 3 weeks', 
                         render.recurrence(date.Recurrence('weekly', amount=3))) 
        
    def testThreeMonthly(self):
        self.assertEqual('Every 3 months', 
                         render.recurrence(date.Recurrence('monthly', 3))) 
        
    def testThreeYearly(self):
        self.assertEqual('Every 3 years', 
                         render.recurrence(date.Recurrence('yearly', 3)))
                
        
class RenderException(test.TestCase):
    def testRenderException(self):
        instance = Exception()
        self.assertEqual(unicode(instance), 
                         render.exception(Exception, instance))

    def testRenderUnicodeDecodeError(self):
        try:
            'abc'.encode('utf-16').decode('utf-8')
        except UnicodeDecodeError, instance:
            self.assertEqual(unicode(instance), 
                             render.exception(UnicodeDecodeError, instance))
            
    def testExceptionThatCannotBePrinted(self):
        """win32all exceptions may contain localized error
        messages. But Exception.__str__ does not handle non-ASCII
        characters in the args instance variable; calling
        unicode(instance) is just like calling str(instance) and
        raises an UnicodeEncodeError."""

        e = Exception(u'é')
        try:
            render.exception(Exception, e)
        except UnicodeEncodeError:  # pragma: no cover
            self.fail() 



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from taskcoachlib import gui, config, persistence, operating_system
from taskcoachlib.domain import attachment


class DummyEvent(object):
    def Skip(self): # pragma: no cover
        pass


class AttachmentEditorTest(test.wxTestCase):
    def setUp(self):
        super(AttachmentEditorTest, self).setUp()
        self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.attachment = attachment.FileAttachment('Attachment')
        self.attachments = attachment.AttachmentList()
        self.attachments.append(self.attachment)
        self.editor = gui.dialog.editor.AttachmentEditor(self.frame, 
            self.attachments, self.settings, self.attachments, self.taskFile)

    def tearDown(self):
        super(AttachmentEditorTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()

    def setSubject(self, newSubject):
        page = self.editor._interior[0]
        page._subjectEntry.SetFocus()
        page._subjectEntry.SetValue(newSubject)
        if operating_system.isGTK(): # pragma: no cover
            page._subjectSync.onAttributeEdited(DummyEvent())
        else: # pragma: no cover
            page._descriptionEntry.SetFocus()
        
    def setDescription(self, newDescription):
        page = self.editor._interior[0]
        page._descriptionEntry.SetFocus()
        page._descriptionEntry.SetValue(newDescription)
        if operating_system.isGTK(): # pragma: no cover
            page._descriptionSync.onAttributeEdited(DummyEvent())
        else: # pragma: no cover
            page._subjectEntry.SetFocus()
        
    def testCreate(self):
        # pylint: disable=W0212
        self.assertEqual('Attachment', self.editor._interior[0]._subjectEntry.GetValue())

    def testEditSubject(self):
        self.setSubject('Done')
        self.assertEqual('Done', self.attachment.subject())

    def testEditDescription(self):
        self.setDescription('Description')
        self.assertEqual('Description', self.attachment.description())
        
    def testAddNote(self):
        viewer = self.editor._interior[1].viewer
        viewer.newItemCommand(viewer.presentation()).do()
        self.assertEqual(1, len(self.attachment.notes())) # pylint: disable=E1101



#! /usr/bin/env python
import wx, sys, os, logging, operator
import wx.lib.newevent
log = logging.getLogger( 'squaremap' )
#log.setLevel( logging.DEBUG )

SquareHighlightEvent, EVT_SQUARE_HIGHLIGHTED = wx.lib.newevent.NewEvent()
SquareSelectionEvent, EVT_SQUARE_SELECTED = wx.lib.newevent.NewEvent()
SquareActivationEvent, EVT_SQUARE_ACTIVATED = wx.lib.newevent.NewEvent()


class HotMapNavigator(object):
    ''' Utility class for navigating the hot map and finding nodes. '''

    @classmethod
    def findNode(class_, hot_map, targetNode, parentNode=None):
        ''' Find the target node in the hot_map. '''
        for index, (rect, node, children) in enumerate(hot_map):
            if node == targetNode:
                return parentNode, hot_map, index
            result = class_.findNode(children, targetNode, node)
            if result:
                return result
        return None

    if hasattr( wx.Rect, 'Contains' ):
        # wx 2.8+
        @classmethod
        def findNodeAtPosition(class_, hot_map, position, parent=None):
            ''' Retrieve the node at the given position. '''
            for rect, node, children in hot_map:
                if rect.Contains(position):
                    return class_.findNodeAtPosition(children, position, node)
            return parent
    else:
        # wx 2.6
        @classmethod
        def findNodeAtPosition(class_, hot_map, position, parent=None):
            ''' Retrieve the node at the given position. '''
            for rect, node, children in hot_map:
                if rect.Inside(position):
                    return class_.findNodeAtPosition(children, position, node)
            return parent

    @staticmethod
    def firstChild(hot_map, index):
        ''' Return the first child of the node indicated by index. '''
        children = hot_map[index][2]
        if children:
            return children[0][1]
        else:
            return hot_map[index][1] # No children, return the node itself

    @staticmethod
    def nextChild(hotmap, index):
        ''' Return the next sibling of the node indicated by index. '''
        nextChildIndex = min(index + 1, len(hotmap) - 1)
        return hotmap[nextChildIndex][1]

    @staticmethod
    def previousChild(hotmap, index):
        ''' Return the previous sibling of the node indicated by index. '''
        previousChildIndex = max(0, index - 1)
        return hotmap[previousChildIndex][1]

    @staticmethod
    def firstNode(hot_map):
        ''' Return the very first node in the hot_map. '''
        return hot_map[0][1]

    @classmethod
    def lastNode(class_, hot_map):
        ''' Return the very last node (recursively) in the hot map. '''
        children = hot_map[-1][2]
        if children:
            return class_.lastNode(children)
        else:
            return hot_map[-1][1] # Return the last node


class SquareMap( wx.Panel ):
    """Construct a nested-box trees structure view"""

    BackgroundColour = wx.Colour( 128,128,128 )
    max_depth = None
    max_depth_seen = None

    def __init__(
        self,  parent=None, id=-1, pos=wx.DefaultPosition,
        size=wx.DefaultSize,
        style=wx.TAB_TRAVERSAL|wx.NO_BORDER|wx.FULL_REPAINT_ON_RESIZE,
        name='SquareMap', model = None,
        adapter = None,
        labels = True, 
        highlight = True,
        padding = 2,
        margin = 0,
        square_style = False,
    ):
        """Initialise the SquareMap
        
        adapter -- a DefaultAdapter or same-interface instance providing SquareMap data api
        labels -- set to True (default) to draw textual labels within the boxes
        highlight -- set to True (default) to highlight nodes on mouse-over 
        padding -- spacing within each square and its children (within the square's border)
        margin -- spacing around each square (on all sides)
        square_style -- use a more-recursive, less-linear, more "square" layout style which 
            works better on objects with large numbers of children, such as Meliae memory 
            dumps, works fine on profile views as well, but the layout is less obvious wrt 
            what node is "next" "previous" etc.
        """
        super( SquareMap, self ).__init__(
            parent, id, pos, size, style, name
        )
        self.model = model
        self.padding = padding
        self.square_style = square_style
        self.margin = margin
        self.labels = labels
        self.highlight = highlight
        self.selectedNode = None
        self.highlightedNode = None
        self._buffer = wx.EmptyBitmap(20, 20) # Have a default buffer ready
        self.Bind( wx.EVT_PAINT, self.OnPaint)
        self.Bind( wx.EVT_SIZE, self.OnSize )
        if highlight:
            self.Bind( wx.EVT_MOTION, self.OnMouse )
        self.Bind( wx.EVT_LEFT_UP, self.OnClickRelease )
        self.Bind( wx.EVT_LEFT_DCLICK, self.OnDoubleClick )
        self.Bind( wx.EVT_KEY_UP, self.OnKeyUp )
        self.hot_map = []
        self.adapter = adapter or DefaultAdapter()
        self.DEFAULT_PEN = wx.Pen( wx.BLACK, 1, wx.SOLID )
        self.SELECTED_PEN = wx.Pen( wx.WHITE, 2, wx.SOLID )
        self.OnSize(None)

    def OnMouse( self, event ):
        """Handle mouse-move event by selecting a given element"""
        node = HotMapNavigator.findNodeAtPosition(self.hot_map, event.GetPosition())
        self.SetHighlight( node, event.GetPosition() )

    def OnClickRelease( self, event ):
        """Release over a given square in the map"""
        node = HotMapNavigator.findNodeAtPosition(self.hot_map, event.GetPosition())
        self.SetSelected( node, event.GetPosition() )

    def OnDoubleClick(self, event):
        """Double click on a given square in the map"""
        node = HotMapNavigator.findNodeAtPosition(self.hot_map, event.GetPosition())
        if node:
            wx.PostEvent( self, SquareActivationEvent( node=node, point=event.GetPosition(), map=self ) )

    def OnKeyUp(self, event):
        event.Skip()
        if not self.selectedNode or not self.hot_map:
            return

        if event.KeyCode == wx.WXK_HOME:
            self.SetSelected(HotMapNavigator.firstNode(self.hot_map))
            return
        elif event.KeyCode == wx.WXK_END:
            self.SetSelected(HotMapNavigator.lastNode(self.hot_map))
            return

        try:
            parent, children, index = HotMapNavigator.findNode(self.hot_map, self.selectedNode)
        except TypeError:
            log.info( 'Unable to find hot-map record for node %s', self.selectedNode )
        else:
            if event.KeyCode == wx.WXK_DOWN:
                self.SetSelected(HotMapNavigator.nextChild(children, index))
            elif event.KeyCode == wx.WXK_UP:
                self.SetSelected(HotMapNavigator.previousChild(children, index))
            elif event.KeyCode == wx.WXK_RIGHT:
                self.SetSelected(HotMapNavigator.firstChild(children, index))
            elif event.KeyCode == wx.WXK_LEFT and parent:
                self.SetSelected(parent)
            elif event.KeyCode == wx.WXK_RETURN:
                wx.PostEvent(self, SquareActivationEvent(node=self.selectedNode,
                                                         map=self))

    def GetSelected(self):
        return self.selectedNode

    def SetSelected( self, node, point=None, propagate=True ):
        """Set the given node selected in the square-map"""
        if node == self.selectedNode:
            return
        self.selectedNode = node
        self.UpdateDrawing()
        if node:
            wx.PostEvent( self, SquareSelectionEvent( node=node, point=point, map=self ) )

    def SetHighlight( self, node, point=None, propagate=True ):
        """Set the currently-highlighted node"""
        if node == self.highlightedNode:
            return
        self.highlightedNode = node
        # TODO: restrict refresh to the squares for previous node and new node...
        self.UpdateDrawing()
        if node and propagate:
            wx.PostEvent( self, SquareHighlightEvent( node=node, point=point, map=self ) )

    def SetModel( self, model, adapter=None ):
        """Set our model object (root of the tree)"""
        self.model = model
        if adapter is not None:
            self.adapter = adapter
        self.UpdateDrawing()

    def OnPaint(self, event):
        dc = wx.BufferedPaintDC(self, self._buffer)

    def OnSize(self, event):
        # The buffer is initialized in here, so that the buffer is always
        # the same size as the Window.
        width, height = self.GetClientSizeTuple()
        if width <= 0 or height <=0:
            return
        # Make new off-screen bitmap: this bitmap will always have the
        # current drawing in it, so it can be used to save the image to
        # a file, or whatever.
        if width and height:
            # Macs can generate events with 0-size values
            self._buffer = wx.EmptyBitmap(width, height)
            self.UpdateDrawing()

    def UpdateDrawing(self):
        dc = wx.BufferedDC(wx.ClientDC(self), self._buffer)
        self.Draw(dc)

    def Draw(self, dc):
        ''' Draw the tree map on the device context. '''
        self.hot_map = []
        dc.BeginDrawing()
        brush = wx.Brush( self.BackgroundColour  )
        dc.SetBackground( brush )
        dc.Clear()
        if self.model:
            self.max_depth_seen = 0
            font = self.FontForLabels(dc)
            dc.SetFont(font)
            self._em_size_ = dc.GetFullTextExtent( 'm', font )[0]
            w, h = dc.GetSize()
            self.DrawBox( dc, self.model, 0,0,w,h, hot_map = self.hot_map )
        dc.EndDrawing()

    def FontForLabels(self, dc):
        ''' Return the default GUI font, scaled for printing if necessary. '''
        font = wx.SystemSettings_GetFont(wx.SYS_DEFAULT_GUI_FONT)
        scale = dc.GetPPI()[0] / wx.ScreenDC().GetPPI()[0]
        font.SetPointSize(scale*font.GetPointSize())
        return font

    def BrushForNode( self, node, depth=0 ):
        """Create brush to use to display the given node"""
        if node == self.selectedNode:
            colour = wx.SystemSettings_GetColour(wx.SYS_COLOUR_HIGHLIGHT)
        elif node == self.highlightedNode:
            colour = wx.Colour( red=0, green=255, blue=0 )
        else:
            colour = self.adapter.background_color(node, depth)
            if not colour:
                red = (depth * 10)%255
                green = 255-((depth * 5)%255)
                blue = (depth * 25)%255
                colour = wx.Colour( red, green, blue )
        return wx.Brush( colour  )

    def PenForNode( self, node, depth=0 ):
        """Determine the pen to use to display the given node"""
        if node == self.selectedNode:
            return self.SELECTED_PEN
        return self.DEFAULT_PEN

    def TextForegroundForNode(self, node, depth=0):
        """Determine the text foreground colour to use to display the label of
           the given node"""
        if node == self.selectedNode:
            fg_colour = wx.SystemSettings_GetColour(wx.SYS_COLOUR_HIGHLIGHTTEXT)
        else:
            fg_colour = self.adapter.foreground_color(node, depth)
            if not fg_colour:
                fg_colour = wx.SystemSettings_GetColour(wx.SYS_COLOUR_WINDOWTEXT)
        return fg_colour

    def DrawBox( self, dc, node, x,y,w,h, hot_map, depth=0 ):
        """Draw a model-node's box and all children nodes"""
        log.debug( 'Draw: %s to (%s,%s,%s,%s) depth %s',
            node, x,y,w,h, depth,
        )
        if self.max_depth and depth > self.max_depth:
            return
        self.max_depth_seen = max( (self.max_depth_seen,depth))
        dc.SetBrush( self.BrushForNode( node, depth ) )
        dc.SetPen( self.PenForNode( node, depth ) )
        # drawing offset by margin within the square...
        dx,dy,dw,dh = x+self.margin,y+self.margin,w-(self.margin*2),h-(self.margin*2)
        if sys.platform == 'darwin':
            # Macs don't like drawing small rounded rects...
            if w < self.padding*2 or h < self.padding*2:
                dc.DrawRectangle( dx,dy,dw,dh )
            else:
                dc.DrawRoundedRectangle( dx,dy,dw,dh, self.padding )
        else:
            dc.DrawRoundedRectangle( dx,dy,dw,dh, self.padding*3 )
#        self.DrawIconAndLabel(dc, node, x, y, w, h, depth)
        children_hot_map = []
        hot_map.append( (wx.Rect( int(x),int(y),int(w),int(h)), node, children_hot_map ) )
        x += self.padding
        y += self.padding
        w -= self.padding*2
        h -= self.padding*2
        

        empty = self.adapter.empty( node )
        icon_drawn = False
        if self.max_depth and depth == self.max_depth:
            self.DrawIconAndLabel(dc, node, x, y, w, h, depth)
            icon_drawn = True
        elif empty:
            # is a fraction of the space which is empty...
            log.debug( '  empty space fraction: %s', empty )
            new_h = h * (1.0-empty)
            self.DrawIconAndLabel(dc, node, x, y, w, h-new_h, depth)
            icon_drawn = True
            y += (h-new_h)
            h = new_h

        if w >self.padding*2 and h> self.padding*2:
            children = self.adapter.children( node )
            if children:
                log.debug( '  children: %s', children )
                self.LayoutChildren( dc, children, node, x,y,w,h, children_hot_map, depth+1 )
            else:
                log.debug( '  no children' )
                if not icon_drawn:
                    self.DrawIconAndLabel(dc, node, x, y, w, h, depth)
        else:
            log.debug( '  not enough space: children skipped' )

    def DrawIconAndLabel(self, dc, node, x, y, w, h, depth):
        ''' Draw the icon, if any, and the label, if any, of the node. '''
        if w-2 < self._em_size_//2 or h-2 < self._em_size_ //2:
            return
        dc.SetClippingRegion(x+1, y+1, w-2, h-2) # Don't draw outside the box
        try:
            icon = self.adapter.icon(node, node==self.selectedNode)
            if icon and h >= icon.GetHeight() and w >= icon.GetWidth():
                iconWidth = icon.GetWidth() + 2
                dc.DrawIcon(icon, x+2, y+2)
            else:
                iconWidth = 0
            if self.labels and h >= dc.GetTextExtent('ABC')[1]:
                dc.SetTextForeground(self.TextForegroundForNode(node, depth))
                dc.DrawText(self.adapter.label(node), x + iconWidth + 2, y+2)
        finally:
            dc.DestroyClippingRegion()
    def LayoutChildren( self, dc, children, parent, x,y,w,h, hot_map, depth=0, node_sum=None ):
        """Layout the set of children in the given rectangle
        
        node_sum -- if provided, we are a recursive call that already has sizes and sorting,
            so skip those operations
        """
        if node_sum is None:
            nodes = [ (self.adapter.value(node,parent),node) for node in children ]
            nodes.sort(key=operator.itemgetter(0))
            total = self.adapter.children_sum( children,parent )
        else:
            nodes = children
            total = node_sum
        if total:
            if self.square_style and len(nodes) > 5:
                # new handling to make parents with large numbers of parents a little less 
                # "sliced" looking (i.e. more square)
                (head_sum,head),(tail_sum,tail) = split_by_value( total, nodes )
                if head and tail:
                    # split into two sub-boxes and render each...
                    head_coord,tail_coord = split_box( head_sum/float(total), x,y,w,h )
                    if head_coord:
                        self.LayoutChildren( 
                            dc, head, parent, head_coord[0],head_coord[1],head_coord[2],head_coord[3],
                            hot_map, depth,
                            node_sum = head_sum,
                        )
                    if tail_coord and coord_bigger_than_padding( tail_coord, self.padding+self.margin ):
                        self.LayoutChildren( 
                            dc, tail, parent, tail_coord[0],tail_coord[1],tail_coord[2],tail_coord[3],
                            hot_map, depth,
                            node_sum = tail_sum,
                        )
                    return 
                        
            (firstSize,firstNode) = nodes[-1]
            fraction = firstSize/float(total)
            head_coord,tail_coord = split_box( firstSize/float(total), x,y,w,h )
            if head_coord:
                self.DrawBox( 
                    dc, firstNode, head_coord[0],head_coord[1],head_coord[2],head_coord[3], 
                    hot_map, depth
                )
            else:
                return # no other node will show up as non-0 either
                
            if len(nodes) > 1 and tail_coord and coord_bigger_than_padding( tail_coord, self.padding+self.margin ):
                self.LayoutChildren( 
                    dc, nodes[:-1], parent, 
                    tail_coord[0],tail_coord[1],tail_coord[2],tail_coord[3], 
                    hot_map, depth,
                    node_sum = total - firstSize,
                )
def coord_bigger_than_padding( tail_coord, padding ):
    return (
        tail_coord and 
        tail_coord[2] > padding * 2 and 
        tail_coord[3] > padding * 2
    )
def split_box( fraction, x,y, w,h ):
    """Return set of two boxes where first is the fraction given"""
    if w >= h:
        new_w = int(w*fraction)
        if new_w:
            return (x,y,new_w,h),(x+new_w,y,w-new_w,h)
        else:
            return None,None
    else:
        new_h = int(h*fraction)
        if new_h:
            return (x,y,w,new_h),(x,y+new_h,w,h-new_h)
        else:
            return None,None
    
def split_by_value( total, nodes, headdivisor=2.0 ):
    """Produce, (sum,head),(sum,tail) for nodes to attempt binary partition"""
    head_sum,tail_sum = 0,0
    divider = 0
    for node in nodes[::-1]:
        if head_sum < total/headdivisor:
            head_sum += node[0]
            divider -= 1
        else:
            break
    return (head_sum,nodes[divider:]),(total-head_sum,nodes[:divider])


class DefaultAdapter( object ):
    """Default adapter class for adapting node-trees to SquareMap API"""
    def children( self, node ):
        """Retrieve the set of nodes which are children of this node"""
        return node.children
    def value( self, node, parent=None ):
        """Return value used to compare size of this node"""
        return node.size
    def label( self, node ):
        """Return textual description of this node"""
        return node.path
    def overall( self, node ):
        """Calculate overall size of the node including children and empty space"""
        return sum( [self.value(value,node) for value in self.children(node)] )
    def children_sum( self, children,node ):
        """Calculate children's total sum"""
        return sum( [self.value(value,node) for value in children] )
    def empty( self, node ):
        """Calculate empty space as a fraction of total space"""
        overall = self.overall( node )
        if overall:
            return (overall - self.children_sum( self.children(node), node))/float(overall)
        return 0
    def background_color(self, node, depth):
        ''' The color to use as background color of the node. '''
        return None
    def foreground_color(self, node, depth):
        ''' The color to use for the label. '''
        return None
    def icon(self, node, isSelected):
        ''' The icon to display in the node. '''
        return None
    def parents( self, node ):
        """Retrieve/calculate the set of parents for the given node"""
        return []


class TestApp(wx.App):
    """Basic application for holding the viewing Frame"""
    def OnInit(self):
        """Initialise the application"""
        wx.InitAllImageHandlers()
        self.frame = frame = wx.Frame( None,
        )
        frame.CreateStatusBar()

        model = model = self.get_model( sys.argv[1])
        self.sq = SquareMap( frame, model=model)
        EVT_SQUARE_HIGHLIGHTED( self.sq, self.OnSquareSelected )
        frame.Show(True)
        self.SetTopWindow(frame)
        return True
    def get_model( self, path ):
        nodes = []
        for f in os.listdir( path ):
            full = os.path.join( path,f )
            if not os.path.islink( full ):
                if os.path.isfile( full ):
                    nodes.append( Node( full, os.stat( full ).st_size, () ) )
                elif os.path.isdir( full ):
                    nodes.append( self.get_model( full ))
        return Node( path, sum([x.size for x in nodes]), nodes )
    def OnSquareSelected( self, event ):
        text = self.sq.adapter.label( event.node )
        self.frame.SetToolTipString( text )

class Node( object ):
    """Really dumb file-system node object"""
    def __init__( self, path, size, children ):
        self.path = path
        self.size = size
        self.children = children
    def __repr__( self ):
        return '%s( %r, %r, %r )'%( self.__class__.__name__, self.path, self.size, self.children )


usage = 'squaremap.py somedirectory'

def main():
    """Mainloop for the application"""
    if not sys.argv[1:]:
        print(usage)
    else:
        app = TestApp(0)
        app.MainLoop()

if __name__ == "__main__":
    main()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import operating_system
from taskcoachlib.thirdparty import customtreectrl as customtree, hypertreelist
from taskcoachlib.widgets import itemctrl, draganddrop
import wx


# pylint: disable=E1101,E1103

class HyperTreeList(draganddrop.TreeCtrlDragAndDropMixin, 
                    hypertreelist.HyperTreeList):
    # pylint: disable=W0223

    def __init__(self, *args, **kwargs):
        super(HyperTreeList, self).__init__(*args, **kwargs)
        if operating_system.isGTK():
            self.Bind(wx.EVT_TREE_ITEM_COLLAPSED, self.__on_item_collapsed)

    def __on_item_collapsed(self, event):
        event.Skip()
        # On Ubuntu, when the user has scrolled to the bottom of the tree
        # and collapses an item, the tree is not redrawn correctly. Refreshing
        # solves this. See http://trac.wxwidgets.org/ticket/11704
        wx.CallAfter(self.MainWindow.Refresh) 

    def GetSelections(self):  # pylint: disable=C0103
        ''' If the root item is hidden, it should never be selected. 
            Unfortunately, CustomTreeCtrl and HyperTreeList allow it to be 
            selected. Override GetSelections to fix that. '''
        selections = super(HyperTreeList, self).GetSelections()
        if self.HasFlag(wx.TR_HIDE_ROOT):
            root_item = self.GetRootItem()
            if root_item and root_item in selections:
                selections.remove(root_item)
        return selections

    def GetMainWindow(self, *args, **kwargs):  # pylint: disable=C0103
        ''' Have a local GetMainWindow so we can create a MainWindow 
        property. '''
        return super(HyperTreeList, self).GetMainWindow(*args, **kwargs)
    
    MainWindow = property(fget=GetMainWindow)
    
    def HitTest(self, point): # pylint: disable=W0221, C0103
        ''' Always return a three-tuple (item, flags, column). '''
        if type(point) == type(()):
            point = wx.Point(point[0], point[1])
        hit_test_result = super(HyperTreeList, self).HitTest(point)
        if len(hit_test_result) == 2:
            hit_test_result += (0,)
        if hit_test_result[0] is None:
            hit_test_result = (wx.TreeItemId(),) + hit_test_result[1:]
        return hit_test_result
    
    def isClickablePartOfNodeClicked(self, event):
        ''' Return whether the user double clicked some part of the node that
            can also receive regular mouse clicks. '''
        return self.__is_collapse_expand_button_clicked(event)
    
    def __is_collapse_expand_button_clicked(self, event):
        flags = self.HitTest(event.GetPosition())[1]
        return flags & wx.TREE_HITTEST_ONITEMBUTTON

    def select(self, selection):
        for item in self.GetItemChildren(recursively=True):
            self.SelectItem(item, self.GetItemPyData(item) in selection)
        
    def clear_selection(self):
        self.UnselectAll()
        self.selectCommand()

    def select_all(self):
        if self.GetItemCount() > 0:
            self.SelectAll()
        self.selectCommand()
                
    def isAnyItemCollapsable(self):
        for item in self.GetItemChildren():
            if self.__is_item_collapsable(item): 
                return True
        return False
    
    def isAnyItemExpandable(self):
        for item in self.GetItemChildren():
            if self.__is_item_expandable(item): 
                return True
        return False
    
    def __is_item_expandable(self, item):
        return self.ItemHasChildren(item) and not self.IsExpanded(item)
    
    def __is_item_collapsable(self, item):
        return self.ItemHasChildren(item) and self.IsExpanded(item)
    
    def IsLabelBeingEdited(self):
        return bool(self.GetLabelTextCtrl())
    
    def StopEditing(self):
        if self.IsLabelBeingEdited():
            self.GetLabelTextCtrl().StopEditing()
            
    def GetLabelTextCtrl(self):
        return self.GetMainWindow()._editCtrl  # pylint: disable=W0212
    
    def GetItemCount(self):
        root_item = self.GetRootItem()
        return self.GetChildrenCount(root_item, recursively=True) \
            if root_item else 0
    

class TreeListCtrl(itemctrl.CtrlWithItemsMixin, itemctrl.CtrlWithColumnsMixin, 
                   itemctrl.CtrlWithToolTipMixin, HyperTreeList):
    # TreeListCtrl uses ALIGN_LEFT, ..., ListCtrl uses LIST_FORMAT_LEFT, ... for
    # specifying alignment of columns. This dictionary allows us to map from the
    # ListCtrl constants to the TreeListCtrl constants:
    alignmentMap = {wx.LIST_FORMAT_LEFT: wx.ALIGN_LEFT, 
                    wx.LIST_FORMAT_CENTRE: wx.ALIGN_CENTRE,
                    wx.LIST_FORMAT_CENTER: wx.ALIGN_CENTER,
                    wx.LIST_FORMAT_RIGHT: wx.ALIGN_RIGHT}
    ct_type = 0
    
    def __init__(self, parent, columns, selectCommand, editCommand, 
                 dragAndDropCommand, itemPopupMenu=None, columnPopupMenu=None, 
                 *args, **kwargs):    
        self.__adapter = parent
        self.__selection = []
        self.__user_double_clicked = False
        self.__columns_with_images = []
        self.__default_font = wx.NORMAL_FONT
        kwargs.setdefault('resizeableColumn', 0)
        super(TreeListCtrl, self).__init__(parent, style=self.__get_style(), 
            agwStyle=self.__get_agw_style(), columns=columns,  
            itemPopupMenu=itemPopupMenu,
            columnPopupMenu=columnPopupMenu, *args, **kwargs)
        self.bindEventHandlers(selectCommand, editCommand, dragAndDropCommand)

    def bindEventHandlers(self, selectCommand, editCommand, dragAndDropCommand):
        # pylint: disable=W0201
        self.selectCommand = selectCommand
        self.editCommand = editCommand
        self.dragAndDropCommand = dragAndDropCommand
        self.Bind(wx.EVT_TREE_SEL_CHANGED, self.onSelect)
        self.Bind(wx.EVT_TREE_KEY_DOWN, self.onKeyDown)
        self.Bind(wx.EVT_TREE_ITEM_ACTIVATED, self.onItemActivated)
        # We deal with double clicks ourselves, to prevent the default behaviour
        # of collapsing or expanding nodes on double click. 
        self.GetMainWindow().Bind(wx.EVT_LEFT_DCLICK, self.onDoubleClick)
        self.Bind(wx.EVT_TREE_BEGIN_LABEL_EDIT, self.onBeginEdit)
        self.Bind(wx.EVT_TREE_END_LABEL_EDIT, self.onEndEdit)
        self.Bind(wx.EVT_TREE_ITEM_EXPANDING, self.onItemExpanding)
        self.Bind(wx.EVT_SET_FOCUS, self.onSetFocus)
        
    def onSetFocus(self, event):  # pylint: disable=W0613
        # Send a child focus event to let the AuiManager know we received focus
        # so it will activate our pane
        wx.PostEvent(self._main_win, wx.ChildFocusEvent(self._main_win))
        self.SetFocus()

    def getItemTooltipData(self, item):
        return self.__adapter.getItemTooltipData(item)
    
    def getItemCTType(self, item): # pylint: disable=W0613
        return self.ct_type
    
    def curselection(self):
        return [self.GetItemPyData(item) for item in self.GetSelections()]
    
    def RefreshAllItems(self, count=0): # pylint: disable=W0613
        self.Freeze()
        self.StopEditing()
        self.__selection = self.curselection()
        self.DeleteAllItems()
        self.__columns_with_images = [index for index in range(self.GetColumnCount()) if self.__adapter.hasColumnImages(index)]
        root_item = self.GetRootItem()
        if not root_item:
            root_item = self.AddRoot('Hidden root')
        self._addObjectRecursively(root_item)
        selections = self.GetSelections()
        if selections:
            self.GetMainWindow()._current = self.GetMainWindow()._key_current = selections[0]
            self.ScrollTo(selections[0])
        self.Thaw()
            
    def RefreshItems(self, *objects):
        self.__selection = self.curselection()
        self._refreshTargetObjects(self.GetRootItem(), *objects)
            
    def _refreshTargetObjects(self, parent_item, *target_objects):
        child_item, cookie = self.GetFirstChild(parent_item)
        while child_item:
            item_object = self.GetItemPyData(child_item) 
            if item_object in target_objects:
                self._refreshObjectCompletely(child_item, item_object)
            self._refreshTargetObjects(child_item, *target_objects)
            child_item, cookie = self.GetNextChild(parent_item, cookie)
            
    def _refreshObjectCompletely(self, item, *args):
        self.__refresh_aspects(('ItemType', 'Columns', 'Font', 'Colors',
                              'Selection'), item, check=True, *args)
        self.GetMainWindow().RefreshLine(item)
        
    def _addObjectRecursively(self, parent_item, parent_object=None):
        for child_object in self.__adapter.children(parent_object):
            child_item = self.AppendItem(parent_item, '', 
                                         self.getItemCTType(child_object), 
                                         data=child_object)
            self._refreshObjectMinimally(child_item, child_object)
            expanded = self.__adapter.getItemExpanded(child_object)
            if expanded:
                self._addObjectRecursively(child_item, child_object)
                # Call Expand on the item instead of on the tree
                # (self.Expand(childItem)) to prevent lots of events
                # (EVT_TREE_ITEM_EXPANDING/EXPANDED) being sent
                child_item.Expand()
            else:
                self.SetItemHasChildren(child_item,
                                        self.__adapter.children(child_object))

    def _refreshObjectMinimally(self, *args, **kwargs):
        self.__refresh_aspects(('Columns', 'Colors', 'Font', 'Selection'), 
                             *args, **kwargs)

    def __refresh_aspects(self, aspects, *args, **kwargs):
        for aspect in aspects:
            refresh_aspect = getattr(self, '_refresh%s' % aspect)
            refresh_aspect(*args, **kwargs)
        
    def _refreshItemType(self, item, domain_object, check=False):
        ct_type = self.getItemCTType(domain_object)
        if not check or (check and ct_type != self.GetItemType(item)):
            self.SetItemType(item, ct_type)
        
    def _refreshColumns(self, item, domain_object, check=False):
        for column_index in range(self.GetColumnCount()):
            self._refreshColumn(item, domain_object, column_index, check=check)
                
    def _refreshColumn(self, item, domain_object, column_index, check=False):
        aspects = ('Text', 'Image') if column_index in self.__columns_with_images else ('Text',)
        self.__refresh_aspects(aspects, item, domain_object, column_index, 
                             check=check)
            
    def _refreshText(self, item, domain_object, column_index, check=False):
        text = self.__adapter.getItemText(domain_object, column_index)
        if text.count('\n') > 3:
            text = '\n'.join(text.split('\n')[:4]) + u' ...'
        if not check or (check and text != item.GetText(column_index)):
            item.SetText(column_index, text)
                
    def _refreshImage(self, item, domain_object, column_index, check=False):
        images = self.__adapter.getItemImages(domain_object, column_index)
        for which, image in images.items():
            image = image if image >= 0 else -1
            if not check or (check and image != item.GetImage(which, 
                                                              column_index)):
                item.SetImage(column_index, image, which)

    def _refreshColors(self, item, domain_object, check=False):
        bg_color = domain_object.backgroundColor(recursive=True) or wx.NullColour
        if not check or (check and bg_color != self.GetItemBackgroundColour(item)):
            self.SetItemBackgroundColour(item, bg_color)
        fg_color = domain_object.foregroundColor(recursive=True) or wx.NullColour
        if not check or (check and fg_color != self.GetItemTextColour(item)):
            self.SetItemTextColour(item, fg_color)
        
    def _refreshFont(self, item, domain_object, check=False):
        font = domain_object.font(recursive=True) or self.__default_font
        if not check or (check and font != self.GetItemFont(item)):
            self.SetItemFont(item, font)
        
    def _refreshSelection(self, item, domain_object, check=False):
        select = domain_object in self.__selection
        if not check or (check and select != item.IsSelected()):
            item.SetHilight(select)

    # Event handlers
    
    def onSelect(self, event):
        # Use CallAfter to prevent handling the select while items are 
        # being deleted:
        wx.CallAfter(self.selectCommand) 
        event.Skip()

    def onKeyDown(self, event):
        if event.GetKeyCode() == wx.WXK_RETURN:
            self.editCommand(event)
        elif event.GetKeyCode() == wx.WXK_F2 and self.GetSelections():
            self.EditLabel(self.GetSelections()[0], column=0)
        else:
            event.Skip()
         
    def OnDrop(self, drop_item, drag_items, part, column):
        drop_item = None if drop_item == self.GetRootItem() else \
                   self.GetItemPyData(drop_item)
        drag_items = list(self.GetItemPyData(drag_item) for drag_item in drag_items)
        wx.CallAfter(self.dragAndDropCommand, drop_item, drag_items, part, column)
        
    def onItemExpanding(self, event):
        event.Skip()
        item = event.GetItem()
        if self.GetChildrenCount(item, recursively=False) == 0:
            domain_object = self.GetItemPyData(item)
            self._addObjectRecursively(item, domain_object)
                
    def onDoubleClick(self, event):
        self.__user_double_clicked = True
        if self.isClickablePartOfNodeClicked(event):
            event.Skip(False)
        else:
            self.onItemActivated(event)
        
    def onItemActivated(self, event):
        ''' Attach the column clicked on to the event so we can use it 
            elsewhere. '''
        column_index = self.__column_under_mouse()
        if column_index >= 0:
            event.columnName = self._getColumn(column_index).name()
        self.editCommand(event)
        event.Skip(False)
        
    def __column_under_mouse(self):
        mouse_position = self.GetMainWindow().ScreenToClient(wx.GetMousePosition())
        item, _, column = self.HitTest(mouse_position)
        if item:
            # Only get the column name if the hittest returned an item,
            # otherwise the item was activated from the menu or by double 
            # clicking on a portion of the tree view not containing an item.
            return max(0, column) # FIXME: Why can the column be -1?
        else:
            return -1
        
    # Inline editing
        
    def onBeginEdit(self, event):
        if self.__user_double_clicked:
            event.Veto()
            self.__user_double_clicked = False
        elif self.IsLabelBeingEdited():
            # Don't start editing another label when the user is still editing
            # a label. This prevents left-over text controls in the tree.
            event.Veto()
        else:
            event.Skip()
        
    def onEndEdit(self, event):
        if event._editCancelled:  # pylint: disable=W0212
            event.Skip()
            return
        event.Veto()  # Let us update the tree
        domain_object = self.GetItemPyData(event.GetItem())
        new_value = event.GetLabel()
        column = self._getColumn(event.GetInt())
        column.onEndEdit(domain_object, new_value)
        
    def CreateEditCtrl(self, item, column_index):
        column = self._getColumn(column_index)
        domain_object = self.GetItemPyData(item)
        return column.editControl(self.GetMainWindow(), item, column_index, 
                                  domain_object)
            
    # Override CtrlWithColumnsMixin with TreeListCtrl specific behaviour:
        
    def _setColumns(self, *args, **kwargs):
        super(TreeListCtrl, self)._setColumns(*args, **kwargs)
        self.SetMainColumn(0)
        for column_index in range(self.GetColumnCount()):
            self.SetColumnEditable(column_index, 
                                   self._getColumn(column_index).isEditable())
                        
    # Extend TreeMixin with TreeListCtrl specific behaviour:

    @staticmethod
    def __get_style():
        return wx.WANTS_CHARS 
    
    @staticmethod        
    def __get_agw_style():
        agw_style = wx.TR_DEFAULT_STYLE | wx.TR_HIDE_ROOT | wx.TR_MULTIPLE \
            | wx.TR_EDIT_LABELS | wx.TR_HAS_BUTTONS | wx.TR_FULL_ROW_HIGHLIGHT \
            | customtree.TR_HAS_VARIABLE_ROW_HEIGHT
        if operating_system.isMac():
            agw_style |= wx.TR_NO_LINES
        agw_style &= ~hypertreelist.TR_NO_HEADER
        return agw_style

    # pylint: disable=W0221
    
    def DeleteColumn(self, column_index):
        self.RemoveColumn(column_index)
        
    def InsertColumn(self, column_index, column_header, *args, **kwargs):
        alignment = self.alignmentMap[kwargs.pop('format', wx.LIST_FORMAT_LEFT)]
        if column_index == self.GetColumnCount():
            self.AddColumn(column_header, *args, **kwargs)
        else:
            super(TreeListCtrl, self).InsertColumn(column_index, column_header, 
                                                   *args, **kwargs)
        self.SetColumnAlignment(column_index, alignment)
        self.SetColumnEditable(column_index, 
                               self._getColumn(column_index).isEditable())

    def showColumn(self, *args, **kwargs):
        ''' Stop editing before we hide or show a column to prevent problems
            redrawing the tree list control contents. '''
        self.StopEditing()
        super(TreeListCtrl, self).showColumn(*args, **kwargs)


class CheckTreeCtrl(TreeListCtrl):
    def __init__(self, parent, columns, selectCommand, checkCommand, 
                 editCommand, dragAndDropCommand, itemPopupMenu=None, 
                 *args, **kwargs):
        self.__checking = False
        super(CheckTreeCtrl, self).__init__(parent, columns,
            selectCommand, editCommand, dragAndDropCommand, 
            itemPopupMenu, *args, **kwargs)
        self.checkCommand = checkCommand
        self.Bind(customtree.EVT_TREE_ITEM_CHECKED, self.onItemChecked)
        self.GetMainWindow().Bind(wx.EVT_LEFT_DOWN, self.onMouseLeftDown)
        self.getIsItemCheckable = parent.getIsItemCheckable if hasattr(parent, 'getIsItemCheckable') else lambda item: True
        self.getIsItemChecked = parent.getIsItemChecked
        self.getItemParentHasExclusiveChildren = parent.getItemParentHasExclusiveChildren
        
    def getItemCTType(self, domain_object):
        ''' Use radio buttons (ct_type == 2) when the object has "exclusive" 
            children, meaning that only one child can be checked at a time. Use
            check boxes (ct_type == 1) otherwise. '''
        if self.getIsItemCheckable(domain_object):
            return 2 if self.getItemParentHasExclusiveChildren(domain_object) else 1
        else:
            return 0
    
    def CheckItem(self, item, checked=True):
        if self.GetItemType(item) == 2:
            # Use UnCheckRadioParent because CheckItem always keeps at least
            # one item selected, which we don't want to enforce
            self.UnCheckRadioParent(item, checked)
        else:
            super(CheckTreeCtrl, self).CheckItem(item, checked)
            
    def onMouseLeftDown(self, event):
        ''' By default, the HyperTreeList widget doesn't allow for unchecking
            a radio item. Since we do want to support unchecking a radio 
            item, we look for mouse left down and uncheck the item and all of
            its children if the user clicks on an already selected radio 
            item. '''
        position = self.GetMainWindow().CalcUnscrolledPosition(event.GetPosition())
        item, flags, dummy_column = self.HitTest(position)
        if item and item.GetType() == 2 and \
           (flags & customtree.TREE_HITTEST_ONITEMCHECKICON) and \
           self.IsItemChecked(item):
            self.__uncheck_item_recursively(item)
        else:
            event.Skip()
            
    def __uncheck_item_recursively(self, item, parent_is_expanded=True, 
                                   disable_item=False):
        if item.GetType():
            self.__uncheck_item(item, torefresh=parent_is_expanded)
        if disable_item:
            self.EnableItem(item, False, torefresh=parent_is_expanded)
        parent_is_expanded = item.IsExpanded()
        child, cookie = self.GetFirstChild(item)    
        while child:
            self.__uncheck_item_recursively(child, parent_is_expanded, 
                                            disable_item=True)
            child, cookie = self.GetNextChild(item, cookie)
            
    def __uncheck_item(self, item, torefresh):
        self.GetMainWindow().CheckItem2(item, checked=False, 
                                        torefresh=torefresh)
        event = customtree.TreeEvent(customtree.wxEVT_TREE_ITEM_CHECKED, 
                                     self.GetId())
        event.SetItem(item)
        event.SetEventObject(self)
        self.GetEventHandler().ProcessEvent(event)
        
    def _refreshObjectCompletely(self, item, domain_object):
        super(CheckTreeCtrl, self)._refreshObjectCompletely(item, domain_object)
        self._refreshCheckState(item, domain_object)
        
    def _refreshObjectMinimally(self, item, domain_object):
        super(CheckTreeCtrl, self)._refreshObjectMinimally(item, domain_object)
        self._refreshCheckState(item, domain_object)
    
    def _refreshCheckState(self, item, domain_object):
        # Use CheckItem2 so no events get sent:
        self.CheckItem2(item, self.getIsItemChecked(domain_object))
        parent = item.GetParent()
        while parent:
            if self.GetItemType(parent) == 2:
                self.EnableItem(item, self.IsItemChecked(parent))
                break
            parent = parent.GetParent()

    def onItemChecked(self, event):
        if self.__checking: 
            # Ignore checked events while we're making the tree consistent,
            # only invoke the callback:
            self.checkCommand(event, final=False)
            return
        self.__checking = True
        item = event.GetItem()
        # Uncheck mutually exclusive children:
        for child in self.GetItemChildren(item):
            if self.GetItemType(child) == 2:
                self.CheckItem(child, False)
                # Recursively uncheck children of mutually exclusive children:
                for grandchild in self.GetItemChildren(child, recursively=True):
                    self.CheckItem(grandchild, False)
        # If this item is mutually exclusive, recursively uncheck siblings 
        # and parent:
        parent = item.GetParent()
        if parent and self.GetItemType(item) == 2:
            for child in self.GetItemChildren(parent):
                if child == item:
                    continue
                self.CheckItem(child, False)
                for grandchild in self.GetItemChildren(child, recursively=True):
                    self.CheckItem(grandchild, False)
            if self.GetItemType(parent) != 2:
                self.CheckItem(parent, False)
        self.__checking = False
        self.checkCommand(event, final=True)
        
    def onItemActivated(self, event):
        if self.__is_double_clicked(event):
            # Invoke super.onItemActivated to edit the item
            super(CheckTreeCtrl, self).onItemActivated(event)
        else:
            # Item is activated, let another event handler deal with the event 
            event.Skip()
    
    @staticmethod        
    def __is_double_clicked(event):
        return hasattr(event, 'LeftDClick') and event.LeftDClick()



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import test
from unittests import dummy
from taskcoachlib import gui, config, persistence, widgets
from taskcoachlib.domain import task
from taskcoachlib.thirdparty.pubsub import pub


class DummyMainWindow(widgets.AuiManagedFrameWithDynamicCenterPane):
    count = 0
    
    def __init__(self):
        super(DummyMainWindow, self).__init__(None)

    def addPane(self, window, caption, floating=False):
        self.count += 1
        super(DummyMainWindow, self).addPane(window, caption, 
                                             str('name%d'%self.count))

    def AddBalloonTip(self, *args, **kwargs):
        pass


class DummyPane(object):
    optionActive = False

    def __init__(self, window):
        self.window = window
        
    def IsToolbar(self):
        return False
    
    def IsNotebookPage(self):
        return True
    
    def IsNotebookControl(self):
        return False
    
    def HasFlag(self, flag):
        return True
    

class DummyEvent(object):
    def __init__(self, pane):
        self._pane = pane
    
    def Skip(self):
        pass

    def GetPane(self):
        return self._pane


class DummyChangeEvent(DummyEvent):
    pass
        

class DummyCloseEvent(DummyEvent):
    def __init__(self, window):
        super(DummyCloseEvent, self).__init__(DummyPane(window))

    
class ViewerContainerTest(test.wxTestCase):
    def setUp(self):
        super(ViewerContainerTest, self).setUp()
        self.events = 0
        task.Task.settings = self.settings = config.Settings(load=False)
        self.settings.set('view', 'viewerwithdummywidgetcount', '2', new=True)
        self.taskFile = persistence.TaskFile()
        self.mainWindow = DummyMainWindow()
        self.container = gui.viewer.ViewerContainer(self.mainWindow,
                                                    self.settings)
        self.viewer1 = self.createViewer('taskviewer1')
        self.container.addViewer(self.viewer1)
        self.viewer2 = self.createViewer('taskviewer2')
        self.container.addViewer(self.viewer2)

    def createViewer(self, settingsSection):
        self.settings.add_section(settingsSection)
        return dummy.ViewerWithDummyWidget(self.mainWindow, self.taskFile, 
            self.settings, settingsSection=settingsSection)
            
    def onEvent(self):
        self.events += 1
    
    def testCreate(self):
        self.assertEqual(0, self.container.size())

    def testAddTask(self):
        self.taskFile.tasks().append(task.Task())
        self.assertEqual(1, self.container.size())

    def testDefaultActiveViewer(self):
        self.assertEqual(self.viewer1, self.container.activeViewer())
        
    def testChangePage_ChangesActiveViewer(self):
        self.container.activateViewer(self.viewer2)
        self.assertEqual(self.viewer2, self.container.activeViewer())

    def testChangePage_NotifiesObserversAboutNewActiveViewer(self):
        pub.subscribe(self.onEvent, 'viewer.status')
        self.container.onPageChanged(DummyChangeEvent(self.viewer2))
        self.failUnless(self.events > 0)
        
    def testCloseViewer_RemovesViewerFromContainer(self):
        self.container.onPageClosed(DummyCloseEvent(self.viewer1))
        self.assertEqual([self.viewer2], self.container.viewers)
        
    def testCloseViewer_ChangesActiveViewer(self):
        self.container.onPageChanged(DummyChangeEvent(self.viewer2))
        self.container.onPageClosed(DummyCloseEvent(self.viewer2))
        self.assertEqual(self.viewer1, self.container.activeViewer())
        
    def testCloseViewer_NotifiesObserversAboutNewActiveViewer(self):
        self.container.activateViewer(self.viewer2)
        pub.subscribe(self.onEvent, 'viewer.status')
        self.container.closeViewer(self.viewer2)
        self.failUnless(self.events > 0)




'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os
import test

# Tests are run with ./tests as current dir, but setup.py expects the project
# root folder to be the current dir. Work around that by changing
# the current dir while importing setup.py:
cwd = os.path.realpath(os.path.curdir)
os.chdir('..')
import setup
os.chdir(cwd)


class EmptFileTest(test.TestCase):
    def emptyPyFiles(self, *folders):
        empty = []
        for root, dirs, files in os.walk(os.path.join(test.projectRoot, *folders)):
            pyfiles = [os.path.join(root, filename) for filename in files 
                       if filename.endswith('.py')]
            for filename in pyfiles:
                if os.stat(filename).st_size == 0:
                    empty.append(filename)
        return empty

    def testNoSourcePyFilesAreEmpty(self):
        self.assertEqual([], self.emptyPyFiles('taskcoachlib'))

    '''
    def testAllUnittestPyFilesAreInManifest(self):
        self.assertEqual([], self.missingPyFiles('tests', 'unittests'))
    
    def testAllReleasetestPyFilesAreInManifest(self):
        self.assertEqual([], self.missingPyFiles('tests', 'releasetests'))

    def testAllIntegrationtestPyFilesAreInManifest(self):
        self.assertEqual([], self.missingPyFiles('tests', 'integrationtests'))
    '''



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import os, time, base


class TestLaunch(base.Win32TestCase):
    def test_launch(self):
        window = self.findWindow(r'^Task Coach$')
        self.failIf(window is None,
                    'Cannot find main window')


class TestWithTaskFile(base.Win32TestCase):
    def setUp(self):
        self.args = ['"%s"'%os.path.join(self.basepath, 'testfile.tsk')]
        super(TestWithTaskFile, self).setUp()

    def test_launch(self):
        self.failUnless(self.findWindow(r'^Task Coach file error$') is None,
                        'Error dialog appeared')
        window = self.findWindow(r'^Task Coach', tries=20)
        self.failIf(window is None,
                    'Cannot find main window')
        self.failUnless(window.title.endswith('testfile.tsk'),
                        'Wrong window title')
        
    def test_save(self):
        filename = self.args[0][1:-1] # Remove "'s
        timestamp = os.stat(filename).st_mtime

        mainwindow = self.findWindow(r'^Task Coach', tries=20)
        w = mainwindow.findChildren('wxWindowClassNR', 'HyperTreeList')
        
        # Double-click the first task to open the task edit dialog:
        for _ in range(2):
            w[1].clickAt(5, 30)
            time.sleep(0.1)

        editor = self.findWindow(r'\(task\)$')
        self.failIf(editor is None, 'Task editor not found')
        editor.waitFocus()

        # Change subject so the task is "dirty":
        editor.sendText(u'New subject')
        # Close the task edit dialog:
        editor.close()

        mainwindow.waitFocus()
        mainwindow.clickAt(58, 15) # Save button

        # Give some time to write the file...
        time.sleep(15)

        if os.path.exists(self.logfilename):
            self.fail('Exception occurred while saving:\n' + \
                      file(self.logfilename, 'rb').read())

        # This fails for a yet unknown reason when launched through the buildbot. Seems
        # to work fine by hand...

        ## self.failUnless(os.stat(filename).st_mtime > timestamp,
        ##                 'File was not written')
        ## self.assertNotEqual(os.path.getsize(filename), 0)



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx
from taskcoachlib import operating_system


if operating_system.isWindows():
    import win32api # pylint: disable=F0401

    class Display(object):
        """
        This class replaces wx.Display on MSW; the original only
        enumerates the displays at app initialization so when people
        start unplugging/replugging monitors things go wrong. Not all
        methods are implemented.
        """
        @staticmethod
        def GetCount():
            return len(win32api.EnumDisplayMonitors(None, None))

        @staticmethod
        def GetFromPoint(p):
            for idx, (_, _, (x1, y1, x2, y2)) in enumerate(win32api.EnumDisplayMonitors(None, None)):
                if p.x >= x1 and p.x < x2 and p.y >= y1 and p.y < y2:
                    return idx
            return wx.NOT_FOUND

        @staticmethod
        def GetFromWindow(window):
            if window.GetWindowStyle() & wx.RESIZE_BORDER:
                margin = wx.SystemSettings.GetMetric(wx.SYS_FRAMESIZE_X)
            else:
                margin = 0

            x, y = window.GetPositionTuple()
            return Display.GetFromPoint(wx.Point(x + margin, y + margin))

        def __init__(self, index):
            self.hMonitor, _, (self.x, self.y, x2, y2) = win32api.EnumDisplayMonitors(None, None)[index]
            self.w = x2 - self.x
            self.h = y2 - self.y

        def GetClientArea(self):
            ns = win32api.GetMonitorInfo(self.hMonitor)
            x1, y1, x2, y2 = ns['Work']
            return wx.Rect(x1, y1, x2 - x1, y2 - y1)

        def GetGeometry(self):
            ns = win32api.GetMonitorInfo(self.hMonitor)
            x1, y1, x2, y2 = ns['Monitor']
            return wx.Rect(x1, y1, x2 - x1, y2 - y1)

        def GetName(self):
            ns = win32api.GetMonitorInfo(self.hMonitor)
            return ns['Device']

        def IsPrimary(self):
            ns = win32api.GetMonitorInfo(self.hMonitor)
            return bool(ns['Flags'] & 1)


    # Monkey-patching so the workaround applies to third party code as
    # well
    wx.Display = Display



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib.domain.category import Category
from taskcoachlib.domain.date import DateTime, TimeDelta
from taskcoachlib.domain.task import Task
from taskcoachlib.i18n import _
from taskcoachlib.thirdparty.dateutil import parser as dparser
import csv
import tempfile
import StringIO
import re
import math


class CSVReader(object):
    def __init__(self, taskList, categoryList):
        self.taskList = taskList
        self.categoryList = categoryList
        
    def createReader(self, fp, dialect, hasHeaders):
        reader = csv.reader(fp, dialect=dialect)
        if hasHeaders:
            reader.next()
        return reader
        
    def read(self, **kwargs):
        fp = tempfile.TemporaryFile()
        fp.write(file(kwargs['filename'], 'rU').read().decode(kwargs['encoding']).encode('UTF-8'))
        fp.seek(0)
        
        rx1 = re.compile(r'^(\d+):(\d+)$')
        rx2 = re.compile(r'^(\d+):(\d+):(\d+)$')

        reader = self.createReader(fp, kwargs['dialect'], kwargs['hasHeaders'])
        dayfirst = kwargs['dayfirst']
        tasksById = dict()
        tasks = []

        for index, line in enumerate(reader):
            if kwargs['importSelectedRowsOnly'] and index not in kwargs['selectedRows']:
                continue
            subject = _('No subject')
            id_ = None
            description = StringIO.StringIO()
            categories = []
            priority = 0
            actualStartDateTime = None
            plannedStartDateTime = None
            dueDateTime = None
            completionDateTime = None
            reminderDateTime = None
            budget = TimeDelta()
            fixedFee = 0.0
            hourlyFee = 0.0
            percentComplete = 0

            for idx, fieldValue in enumerate(line):
                if kwargs['mappings'][idx] == _('ID'):
                    id_ = fieldValue.decode('UTF-8')
                elif kwargs['mappings'][idx] == _('Subject'):
                    subject = fieldValue.decode('UTF-8')
                elif kwargs['mappings'][idx] == _('Description'):
                    description.write(fieldValue.decode('UTF-8'))
                    description.write(u'\n')
                elif kwargs['mappings'][idx] == _('Category') and fieldValue:
                    name = fieldValue.decode('UTF-8')
                    if name.startswith('(') and name.endswith(')'):
                        continue  # Skip categories of subitems
                    cat = self.categoryList.findCategoryByName(name)
                    if not cat:
                        cat = self.createCategory(name)
                    categories.append(cat)
                elif kwargs['mappings'][idx] == _('Priority'):
                    try:
                        priority = int(fieldValue)
                    except ValueError:
                        pass
                elif kwargs['mappings'][idx] == _('Actual start date'):
                    actualStartDateTime = self.parseDateTime(fieldValue, dayfirst=dayfirst)
                elif kwargs['mappings'][idx] == _('Planned start date'):
                    plannedStartDateTime = self.parseDateTime(fieldValue, dayfirst=dayfirst)
                elif kwargs['mappings'][idx] == _('Due date'):
                    dueDateTime = self.parseDateTime(fieldValue, 23, 59, 59, dayfirst=dayfirst) 
                elif kwargs['mappings'][idx] == _('Completion date'):
                    completionDateTime = self.parseDateTime(fieldValue, 12, 0, 0, dayfirst=dayfirst) 
                elif kwargs['mappings'][idx] == _('Reminder date'):
                    reminderDateTime = self.parseDateTime(fieldValue, dayfirst=dayfirst)
                elif kwargs['mappings'][idx] == _('Budget'):
                    try:
                        value = float(fieldValue)
                        hours = int(math.floor(value))
                        minutes = int(60 * (value - hours))
                        budget = TimeDelta(hours=hours, minutes=minutes, seconds=0)
                    except ValueError:
                        mt = rx1.search(fieldValue)
                        if mt:
                            budget = TimeDelta(hours=int(mt.group(1)), minutes=int(mt.group(2)), seconds=0)
                        else:
                            mt = rx2.search(fieldValue)
                            if mt:
                                budget = TimeDelta(hours=int(mt.group(1)), minutes=int(mt.group(2)), seconds=int(mt.group(3)))
                elif kwargs['mappings'][idx] == _('Fixed fee'):
                    try:
                        fixedFee = float(fieldValue)
                    except ValueError:
                        pass
                elif kwargs['mappings'][idx] == _('Hourly fee'):
                    try:
                        hourlyFee = float(fieldValue)
                    except ValueError:
                        pass
                elif kwargs['mappings'][idx] == _('Percent complete'):
                    try:
                        percentComplete = max(0, min(100, int(fieldValue)))
                    except ValueError:
                        pass

            task = Task(subject=subject,
                        description=description.getvalue(),
                        priority=priority,
                        actualStartDateTime=actualStartDateTime,
                        plannedStartDateTime=plannedStartDateTime,
                        dueDateTime=dueDateTime,
                        completionDateTime=completionDateTime,
                        reminder=reminderDateTime,
                        budget=budget,
                        fixedFee=fixedFee,
                        hourlyFee=hourlyFee,
                        percentageComplete=percentComplete)

            if id_ is not None:
                tasksById[id_] = task

            for category in categories:
                category.addCategorizable(task)
                task.addCategory(category)

            tasks.append(task)

        # OmniFocus uses the task's ID to keep track of hierarchy: 1 => 1.1 and 1.2, etc...

        if tasksById:
            ids = []
            for id_, task in tasksById.items():
                try:
                    ids.append(tuple(map(int, id_.split('.'))))
                except ValueError:
                    self.taskList.append(task)

            ids.sort()
            ids.reverse()

            for id_ in ids:
                sid = '.'.join(map(str, id_))
                if len(id_) >= 2:
                    pid = '.'.join(map(str, id_[:-1]))
                    if pid in tasksById:
                        tasksById[pid].addChild(tasksById[sid])
                else:
                    self.taskList.append(tasksById[sid])
        else:
            self.taskList.extend(tasks)

    def createCategory(self, name):
        if ' -> ' in name:
            parentName, childName = name.rsplit(' -> ', 1)
            parent = self.categoryList.findCategoryByName(parentName)
            if not parent:
                parent = self.createCategory(parentName)
            newCategory = Category(subject=childName)
            parent.addChild(newCategory)
            newCategory.setParent(parent)
        else:
            newCategory = Category(subject=name)
        self.categoryList.append(newCategory)
        return newCategory

    def parseDateTime(self, fieldValue, defaultHour=0, defaultMinute=0, 
                      defaultSecond=0, dayfirst=False):
        if not fieldValue:
            return None
        try:
            dateTime = dparser.parse(fieldValue.decode('UTF-8'), 
                                     dayfirst=dayfirst, fuzzy=True).replace(tzinfo=None)
            hour, minute, second = dateTime.hour, dateTime.minute, dateTime.second
            if 0 == hour == minute == second:
                hour = defaultHour
                minute = defaultMinute
                second = defaultSecond
            return DateTime(dateTime.year, dateTime.month, dateTime.day,
                            hour, minute, second)
        except:  # pylint: disable=W0702
            return None  



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import wx

class wxTimeFormat(object):
	"""Singleton used to format times."""

	def __init__(self):
		def defaultFormat(dt, includeMinutes=False):
			if includeMinutes:
				return dt.Format('%H:%M')
			return dt.Format('%H')
		self.__fmt = defaultFormat

	def FormatTime(self, dateTime, includeMinutes=False):
		return self.__fmt(dateTime, includeMinutes=includeMinutes)

	def SetFormatFunction(self, func):
		self.__fmt = func

wxTimeFormat = wxTimeFormat()
